from multilogging import multilogger
from typing import List

import abc
import re
import random


ZERO = 0
EMPTY_STRING = ""
EMPTY_LIST = []

logger = multilogger(name="DicerGirl", payload="Dicer")


class BaseDice:
    def __init__(self, roll_string: str = EMPTY_STRING) -> None:
        self.roll_string = roll_string
        self.db = EMPTY_STRING
        self.outcome = ZERO
        self.display = EMPTY_LIST

    def __repr__(self) -> str:
        return self.db.upper()

    @abc.abstractmethod
    def parse(self) -> "BaseDice":
        raise NotImplementedError

    @abc.abstractmethod
    def roll(self) -> int:
        """对骰子进行投掷并给出结果"""
        raise NotImplementedError


class DigitDice(BaseDice):
    """数字骰"""

    def __init__(self, roll_string: str = EMPTY_STRING) -> None:
        super().__init__(roll_string=roll_string)
        if not roll_string.isdigit():
            raise ValueError

        self.parse()

    def parse(self) -> "DigitDice":
        self.a = int(self.roll_string)
        self.b = 1
        self.db = f"{self.a}"
        return self

    def roll(self) -> int:
        self.outcome = self.a
        self.display = [self.a]
        return self.outcome


class Dice(BaseDice):
    """多面骰"""

    def __init__(self, roll_string: str = "", explode: bool = False) -> None:
        super().__init__(roll_string=roll_string)
        self.dices = EMPTY_LIST
        self.great = False
        self.explode = explode
        self.parse()

    def parse(self) -> "Dice":
        self.dices = []
        split = re.split(r"[dD]", self.roll_string)

        if split[0]:
            self.a = int(split[0])
        else:
            self.a = 1

        if split[1]:
            self.b = int(split[1])
        else:
            self.b = 100

        self.db = f"{self.a}D{self.b}"
        self.dices += [f"D{self.b}"] * self.a
        return self

    def roll(self) -> int:
        self.results = []
        self.display = []

        for _ in range(self.a):
            result = random.randint(1, self.b)

            if result == 1 and self.explode:
                result -= 1

            if self.explode and self.b == 8:
                self.dices.append("D10")
                result2 = random.randint(1, 10)
                if result2 == 1:
                    result -= 1
                result += result2
                if result2 == 10:
                    self.dices.append("D12")
                    result3 = random.randint(1, 12)
                    if result3 == 1:
                        result -= 1
                    result += result3
                    if result3 == 12:
                        self.dices.append("D20")
                        result4 = random.randint(1, 20)
                        if result4 == 1:
                            result -= 1
                        result += result4
                        if result4 == 20:
                            self.great = True

            self.results.append(result)
            self.display.append(result)

        self.outcome = sum(self.results)
        return self.outcome


class AwardDice(BaseDice):
    """奖励骰"""

    def __init__(self, roll_string: str = "") -> None:
        super().__init__(roll_string=roll_string)
        self.parse()

    def parse(self) -> "AwardDice":
        split = re.split(r"[bB]", self.roll_string)

        if split[0]:
            self.a = int(split[0])
        else:
            self.a = 1

        self.b = int(split[1])
        self.db = f"{self.a}B{self.b}"
        return self

    def roll(self) -> int:
        self.results = []
        self.display = []

        for _ in range(self.a):
            ten = []
            for _ in range(self.b):
                outcome = Dice("1d10").roll()
                outcome = outcome if outcome != 10 else 0
                ten.append(outcome)

            result = Dice("1d100").roll()
            ten.append(result // 10)
            minten = min(ten)
            ten.remove(result // 10)
            outcome = minten * 10 + (result % 10)
            self.results.append(outcome)
            self.display.append([result, ten])

        self.outcome = sum(self.results)
        return self.outcome


class PunishDice(BaseDice):
    """惩罚骰"""

    def __init__(self, roll_string: str = "") -> None:
        super().__init__(roll_string=roll_string)
        self.parse()

    def parse(self) -> "PunishDice":
        split = re.split(r"[pP]", self.roll_string)

        if split[0]:
            self.a = int(split[0])
        else:
            self.a = 1

        self.b = int(split[1])
        self.db = f"{self.a}P{self.b}"
        return self

    def roll(self) -> int:
        self.results = []
        self.display = []

        for _ in range(self.a):
            ten = []
            for _ in range(self.b):
                outcome = Dice("1d10").roll()
                outcome = outcome if outcome != 10 else 0
                ten.append(outcome)

            result = Dice("1d100").roll()
            ten.append(result // 10)
            maxten = max(ten)
            ten.remove(result // 10)
            outcome = maxten * 10 + (result % 10)
            self.results.append(outcome)
            self.display.append([result, ten])

        self.outcome = sum(self.results)
        return self.outcome


class Dicer:
    """掷骰类
    参数:
        roll_string: 标准掷骰表达式
        explode: 是否启用爆炸骰
    示例:
        ```python
        dice = Dice("1d10")
        dice.roll()
        print(dice.outcome) # 输出`1d10`投掷结果
        ```
    """

    def __init__(self, roll_string: str = EMPTY_STRING, explode: bool = False) -> None:
        self.roll_string: str = roll_string
        self.explode: bool = explode
        self.calc_list: List[str | Dice | DigitDice | AwardDice | PunishDice] = []
        self.results: List[int] = []
        self.display: List[int | List[int]] = []
        self.outcome: int = ZERO
        self.great: bool = False
        self.dices: List[str] = []

    def parse(self, roll_string: str = EMPTY_STRING, explode: bool = False):
        self.roll_string = roll_string if roll_string else self.roll_string
        self.calc_list = []
        self.db = EMPTY_STRING
        matches: List[str] = re.findall(r"\d*[a-zA-Z]\w*|\d+|[-+*/]", self.roll_string)

        for match in matches:
            if match in ("+", "-", "*", "/", "(", ")"):
                self.calc_list.append(match)
                self.db += match
            elif re.match(r"\d*[dD]\d*", match):
                self.calc_list.append(Dice(match, explode=explode))
                self.db += match.upper()
            elif re.match(r"\d*[bB]\d+", match):
                self.calc_list.append(AwardDice(match))
                self.db += match.upper()
            elif re.match(r"\d*[pP]\d+", match):
                self.calc_list.append(PunishDice(match))
                self.db += match.upper()
            elif re.match(r"\d+", match):
                self.calc_list.append(DigitDice(match))
                self.db += match.upper()
            else:
                raise ValueError(f"骰 {match} 不符合规范.")

        if not matches:
            self.calc_list.append(Dice("1d100"))
            self.db = "1D100"

        return self

    def roll(self):
        self.parse(roll_string=self.roll_string, explode=self.explode)
        self.dices = []
        self.display = []
        for index, calc in enumerate(self.calc_list):
            if calc in ("+", "-", "*", "/", "(", ")"):
                continue

            outcome = calc.roll()
            self.calc_list[index] = outcome
            self.results.append(outcome)
            self.display += calc.display

            if isinstance(calc, Dice) and self.explode:
                if calc.great:
                    self.great = True

                self.dices += calc.dices

        self.outcome = eval("".join(map(str, self.calc_list)))
        return self

    def description(self):
        def count_integers(lst) -> int:
            count = 0
            for item in lst:
                if isinstance(item, int):
                    count += 1
                elif isinstance(item, list):
                    count += count_integers(item)
            return count

        results = self.display
        len_display = count_integers(self.display)
        len_results = count_integers(self.results)

        if len_display <= 10:
            results = self.display
        elif len_results <= 10:
            results = self.results
        else:
            results = [...]

        return f"{self.db}={results}={self.outcome}"

    def get_results(self):
        return self.results

    def detail_expr(self):
        return str(self.results)

    def calc(self):
        return self.outcome

    def __repr__(self):
        return self.db


if __name__ == "__main__":
    # text = "-10/d2/1d10+2d2-22/2+3p2+2b10-p4/b2/d2"
    # dice = Dicer(text)
    # print(dice.calc_list)
    # dice.roll()
    # print(dice.calc_list)
    # print(dice.results)
    # print(dice.outcome)
    roll_strings = {
        "1": 1,
        "10": 10,
        "100": 100,
        "-1": -1,
        "-10": -10,
        "-100": -100,
        "1d1": 1,
        "10d1": 10,
        "100d1": 100,
        "10d1+10d1": 20,
        "10d1-10d1": 0,
        "10d1+10d1+10d1": 30,
        "10d1-10d1+10d1": 10,
        "10d1-10d1-10d1": -10,
    }
    for roll_string in roll_strings.keys():
        try:
            dice = Dicer().parse(roll_string).roll().roll()
            if dice.outcome != roll_strings[roll_string]:
                print(dice.description())
                raise ValueError(
                    f"对于 {roll_string} dice.toal={dice.outcome} 但期待 {roll_strings[roll_string]}"
                )
        except ValueError as error:
            logger.exception(error)

    try:
        roll_string = "d"
        dice = Dicer(roll_string).roll()
        print(dice.description())
    except ValueError as error:
        logger.exception(error)



import os
from enum import unique, Enum
from typing import List, Dict

from multilogging import multilogger
from ruamel.yaml import CommentedMap, CommentedSeq

from dicergirl.common import const
from dicergirl.common.const import DEFAULT_GROUP_NAME as NAME
from dicergirl.common.response import GenericResponse, ConditionResponse
from dicergirl.reply.data import GenericData, ConditionData
from dicergirl.reply.parsers.matcher import MatchType

logger = multilogger(name="DicerGirl", payload="ReplyRegistry")


@unique
class ReplyType(Enum):
    """
    事件类型

    - DEFAULT: 默认回复事件，对应值为0。
    - CUSTOM: 自定义回复事件，对应值为1。
    - CONDITION: 条件回复事件，对应值为2。
    """

    DEFAULT = (0,)
    CUSTOM = (1,)
    CONDITION = 2


class ReplyRegistry:
    """
    ReplyRegistry 负责存储所有需要交由用户自定义的事件。

    通用回复和特定条件回复的自定义无需手动注册，请使用 dicergirl.reply.init 下的 init 方法。

    Attributes:
        _instance (ReplyRegistry): ReplyRegistry 的实例对象。
        _default_generic_data (dict): 存储默认回复的字典。
        _custom_generic_data (dict): 存储自定义通用回复的字典。
        _condition_specific_data (dict): 存储条件回复的字典。

    Methods:
        register_container: 注册数据到相应的容器中
        register_event: 注册事件
        remove_event: 销毁事件
        enable_event: 启用事件
        disable_event: 禁用事件
        toggle_event: 切换事件状态
        generic_event_names: 获取所有默认事件的事件
        custom_event_names: 获取所有自定义事件的事件名
        message_event_names: 获取所有消息事件的事件名
        all_event_names: 获取所有事件的事件名

    Private Methods:
        _register_default_generic_event: 注册默认回复事件
        _register_custom_generic_event: 注册自定义通用回复事件
        _register_condition_specific_event: 注册条件回复事件
        _remove_default_generic_event: 从默认通用数据容器中删除指定事件
        _remove_custom_generic_event: 从自定义通用数据容器中删除指定事件，并从文件中擦除相关数据
        _remove_condition_specific_event: 从条件数据容器中删除指定事件，并从文件中擦除相关数据
        _add_data_in_file: 将数据添加到文件中
        _erase_data_in_file: 从文件中擦除数据

    Usage:
        registry.register_event("common.sayhi", "你好{user}!")  # 注册默认通用回复
        registry.register_event("common.sayhi", "你好{user}!现在是北京时间{time}!", is_custom=True)  # 注册自定义通用回复
        registry.register_event("common.sayhi", "你好{user}！", "你好", MatchType.EXACT_MATCH)  # 注册特定条件下的自定义回复
    """

    _instance = None
    _default_generic_data: Dict[str, GenericData] = {}
    _custom_generic_data: Dict[str, GenericData] = {}
    _condition_specific_data: Dict[str, ConditionData] = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._default_generic_data[NAME] = GenericData(NAME, "0.1")
            cls._custom_generic_data[NAME] = GenericData(f"dg-{NAME}", "0.1")
            cls._condition_specific_data[NAME] = ConditionData(NAME, "0.1")
            cls._instance = object.__new__(cls)
        return cls._instance

    def disable_event(
        self, event_name: str, is_message_event: bool = False, group_name: str = NAME
    ):
        """
        禁用事件

        Args:
            event_name (str): 待禁用事件的事件名
            is_message_event (bool, optional): 是否为消息事件，用于标识特定条件下的回复事件。默认值为 False。
            group_name (str, optional): 数据组的名称。默认值为 NAME。

        Returns:
            bool: 当前事件的状态
        """
        try:
            if is_message_event:
                container = self._condition_specific_data[NAME]
                filename = f"{group_name}.yml"
                cache = const.CONDITION_SPECIFIC_REPLY_FILE_CACHE
            else:
                container = self._custom_generic_data[f"dg-{group_name}"]
                filename = f"dg-{group_name}.yml"
                cache = const.GENERIC_REPLY_FILE_CACHE
            response = container.disable(event_name)
            if not response:
                logger.info(f"没有找到{event_name}对应的Response实例")
                return None
            self._add_data_in_file(filename, cache, response, True)
            return container.is_enable(event_name)
        except KeyError as e:
            logger.error(f"请确保您的回复配置文件包含了正确的键和相应的值。如果您不确定如何正确配置文件，请参考文档或向管理员寻求帮助。")
            logger.error(f"Error: {e}")
            return None

    def enable_event(
        self, event_name: str, is_message_event: bool = False, group_name: str = NAME
    ):
        """
        启用事件

        Args:
            event_name (str): 待启用事件的事件名
            is_message_event (bool, optional): 是否为消息事件，用于标识特定条件下的回复事件。默认值为 False。
            group_name (str, optional): 数据组的名称。默认值为 NAME。
        Returns:
            bool: 当前事件的状态
        """
        try:
            if is_message_event:
                container = self._condition_specific_data[group_name]
                filename = f"{group_name}.yml"
                cache = const.CONDITION_SPECIFIC_REPLY_FILE_CACHE
            else:
                container = self._custom_generic_data[f"dg-{group_name}"]
                filename = f"dg-{group_name}.yml"
                cache = const.GENERIC_REPLY_FILE_CACHE
            response = container.enable(event_name)
            if not response:
                logger.info(f"没有找到{event_name}对应的Response实例")
                return None
            self._add_data_in_file(filename, cache, response, True)
            return container.is_enable(event_name)
        except KeyError as e:
            logger.error(f"请确保您的回复配置文件包含了正确的键和相应的值。如果您不确定如何正确配置文件，请参考文档或向管理员寻求帮助。")
            logger.error(f"Error: {e}")
            return None

    def toggle(
        self, event_name: str, is_message_event: bool = False, group_name: str = NAME
    ):
        """
        切换事件状态

        Args:
            event_name (str): 待切换的事件名
            is_message_event (bool, optional): 是否为消息事件，用于标识特定条件下的回复事件。默认值为 False。
            group_name (str, optional): 数据组的名称。默认值为 NAME。
        Returns:
            bool: 当前事件的状态
        """
        try:
            if is_message_event:
                container = self._condition_specific_data[group_name]
                filename = f"{group_name}.yml"
                cache = const.CONDITION_SPECIFIC_REPLY_FILE_CACHE
            else:
                container = self._custom_generic_data[f"dg-{group_name}"]
                filename = f"dg-{group_name}.yml"
                cache = const.GENERIC_REPLY_FILE_CACHE
            response = container.toggle(event_name)
            if not response:
                logger.info(f"没有找到{event_name}对应的Response实例")
                return None
            self._add_data_in_file(filename, cache, response, True)
            return container.is_enable(event_name)
        except KeyError as e:
            logger.error(f"请确保您的回复配置文件包含了正确的键和相应的值。如果您不确定如何正确配置文件，请参考文档或向管理员寻求帮助。")
            logger.error(f"Error: {e}")
            return None

    def register_container(self, data: GenericData) -> bool:
        """
        注册数据到相应的容器中

        Args:
            data (GenericData): 要注册的数据对象

        Returns:
            bool: 如果注册成功，则返回 True；否则返回 False。
        """
        if isinstance(data, ConditionData):
            container = self._condition_specific_data
        elif isinstance(data, GenericData):
            container = self._custom_generic_data
        else:
            logger.warning(f"回复文件'{data.group_name.upper()}'注册失败")
            return False
        container[data.group_name] = data
        if data.group_name in container:
            logger.debug(f"自定义回复文件'{data.group_name.upper()}'注册成功")
            return True
        else:
            logger.warning(f"回复文件'{data.group_name.upper()}'注册失败")
            return False

    def register_event(
        self,
        event_name: str,
        send_text: str,
        match_field: str = None,
        match_type: MatchType = None,
        enable: bool = True,
        is_custom: bool = False,
    ) -> bool:
        """
        注册事件:
        若只向 event_name, send_text 传参，则向 _default_generic_data 注册。
        若 match_field 和 match_type 不为空, 则向 _condition_specific_data 注册。
        若以上条件不皆满足, 且自定义标识为 True, 则向 _custom_generic_data 注册。

        Args:
            event_name (str): 事件名
            send_text (str): 事件对应的响应文本
            match_field (str): 匹配字段，例如: "<[^>]*>"
            match_type (MatchType): 匹配类型:相等、包含、正则、函数
            enable (bool): 是否启用响应事件
            is_custom (bool): 标记，是否为用户自定义事件

        Returns:
            bool: 如果事件成功注册到相关容器中，则返回 True；否则返回 False。
        """
        if match_field and match_type:
            return self._register_condition_specific_event(
                event_name, send_text, match_field, match_type, enable
            )
        elif is_custom:
            return self._register_custom_generic_event(event_name, send_text, enable)
        else:
            return self._register_default_generic_event(event_name, send_text, enable)

    def remove_event(
        self,
        event_name: str,
        group_name: str = NAME,
        reply_type: ReplyType = ReplyType.CUSTOM,
    ) -> bool:
        """
        销毁事件

        Args:
            event_name (str): 要删除的事件的名称。
            group_name (str, optional): 数据组的名称。默认值为 NAME。
            reply_type (ReplyType, optional): 事件类型。默认值为 ReplyType。

        Returns:
            bool: 如果成功删除事件，则返回 True；否则返回 False。
        """
        if reply_type is ReplyType.CUSTOM:
            return self._remove_custom_generic_event(event_name, f"dg-{group_name}")
        elif reply_type is ReplyType.CONDITION:
            return self._remove_condition_specific_event(event_name, group_name)
        elif reply_type is ReplyType.DEFAULT:
            return self._remove_default_generic_event(event_name, group_name)
        else:
            logger.error(f"错误的回复类型: {str(reply_type)}")
            return False

    def _register_default_generic_event(
        self, event_name: str, send_text: str, enable: bool = True
    ) -> bool:
        """
        注册默认回复事件

        Args:
            event_name (str):  事件名
            send_text (str): 事件对应的响应文本
            enable (bool): 是否启用事件

        Returns:
            bool: 如果事件成功注册到_default_generic_data中，则返回 True；否则返回 False。
        """
        container = self._default_generic_data[NAME]
        response_type = "DicerGirl默认回复"

        response = GenericResponse(event_name, send_text, enable)
        container.add(response)

        if container.get_response(event_name):
            logger.debug(f"{response_type}'{event_name}'注册成功")
            return True
        else:
            logger.warning(f"{response_type}'{event_name}'注册失败")
            return False

    def _register_custom_generic_event(
        self, event_name: str, send_text: str, enable: bool = True
    ) -> bool:
        """
        注册自定义通用回复事件

        Args:
            event_name (str):  事件名
            send_text (str): 事件对应的响应文本
            enable (bool): 是否启用事件

        Returns:
            bool: 如果事件成功注册到_custom_generic_data中，则返回 True；否则返回 False。
        """
        container = self._custom_generic_data[f"dg-{NAME}"]
        response_type = "自定义通用回复"
        filename = f"dg-{NAME}.yml"
        cache = const.GENERIC_REPLY_FILE_CACHE

        response = GenericResponse(event_name, send_text, enable)
        container.add(response)

        self._add_data_in_file(filename, cache, response)

        if container.get_response(event_name):
            logger.debug(f"{response_type}'{event_name}'注册成功")
            return True
        else:
            logger.warning(f"{response_type}'{event_name}'注册失败")
            return False

    def _register_condition_specific_event(
        self,
        event_name: str,
        send_text: str,
        match_field: str,
        match_type: MatchType,
        enable: bool = True,
    ) -> bool:
        """
        注册条件回复事件

        Args:
            event_name (str): 事件名
            send_text (str): 事件对应的响应文本
            match_field (str): 匹配字段，例如: "<[^>]*>"
            match_type (MatchType): 匹配类型:相等、包含、正则、函数
            enable (bool): 是否启用响应事件

        Returns:
            bool: 如果事件成功注册到_condition_specific_data中，则返回 True；否则返回 False。
        """
        container = self._condition_specific_data[NAME]
        response_type = "自定义条件回复"
        filename = f"{NAME}.yml"
        cache = const.CONDITION_SPECIFIC_REPLY_FILE_CACHE

        response = ConditionResponse(
            event_name, send_text, match_field, match_type, enable
        )
        container.add(response)

        self._add_data_in_file(filename, cache, response)

        if container.get_response(event_name):
            logger.debug(f"{response_type}'{event_name}'注册成功")
            return True
        else:
            logger.warning(f"{response_type}'{event_name}'注册失败")
            return False

    def _remove_default_generic_event(self, event_name, group_name: str = NAME):
        """
        从默认通用数据容器中删除指定事件。

        Args:
            event_name (str): 要删除的事件的名称。
            group_name (str, optional): 数据组的名称。默认值为 NAME。

        Returns:
            bool: 如果成功删除事件，则返回 True；否则返回 False。
        """
        container = self._default_generic_data[group_name]
        container.remove(event_name)
        if not container.get_response(event_name):
            logger.debug(f"回复事件'{event_name}'销毁成功")
            return True
        else:
            logger.warning(f"回复事件'{event_name}'销毁失败")
            return False

    def _remove_custom_generic_event(self, event_name, group_name: str = NAME):
        """
        从自定义通用数据容器中删除指定事件，并从文件中擦除相关数据。

        Args:
            event_name (str): 要删除的事件的名称。
            group_name (str, optional): 数据组的名称。默认值为 NAME。

        Returns:
            bool: 如果成功删除事件，则返回 True；否则返回 False。
        """
        filename = f"{group_name}.yml"
        cache = const.GENERIC_REPLY_FILE_CACHE

        container = self._custom_generic_data[group_name]
        container.remove(event_name)
        self.erase_data_in_file(filename, event_name, cache)
        if not container.get_response(event_name):
            logger.debug(f"回复事件'{event_name}'销毁成功")
            return True
        else:
            logger.warning(f"回复事件'{event_name}'销毁失败")
            return False

    def _remove_condition_specific_event(self, event_name, group_name: str = NAME):
        """
        从条件数据容器中删除指定事件，并从文件中擦除相关数据。

        Args:
            event_name (str): 要删除的事件的名称。
            group_name (str, optional): 数据组的名称。默认值为 NAME。

        Returns:
            bool: 如果成功删除事件，则返回 True；否则返回 False。
        """
        filename = f"{group_name}.yml"
        cache = const.CONDITION_SPECIFIC_REPLY_FILE_CACHE

        container = self._condition_specific_data[group_name]
        container.remove(event_name)

        self.erase_data_in_file(filename, event_name, cache)
        self.erase_data_in_file(filename, event_name, cache)

        if not container.get_response(event_name):
            logger.debug(f"回复事件'{event_name}'销毁成功")
            return True
        else:
            logger.warning(f"回复事件'{event_name}'销毁失败")
            return False

    @staticmethod
    def _add_data_in_file(
        filename, cache, response: GenericResponse, is_update=False
    ) -> bool:
        """
        将数据添加到文件中。

        Args:
            filename (str): 文件名
            cache (dict): 缓存数据的字典
            response (GenericResponse): 响应对象

        Returns:
            bool: 如果成功将数据添加到文件中，则返回 True；否则返回 False。

        Raises:
            Exception: 捕获读写YAML文件时产生的未知异常。
        """
        try:
            event_name = response.event_name
            for filepath, data in cache.items():
                if os.path.basename(filepath) == filename:
                    save_data = response.to_dict()
                    tmp = data["items"]
                    if isinstance(tmp, CommentedSeq):
                        found = False
                        for item in list(tmp):
                            if isinstance(item, CommentedMap):
                                for name, content in item.items():
                                    if name == event_name:
                                        item[name] = save_data
                                        found = True
                                        break
                        if not found and not is_update:
                            data["items"].append({event_name: save_data})
                        with open(file=filepath, mode="wb") as drf:
                            const.REPLY_YAML.dump(data=data, stream=drf)
                            break
        except Exception as e:
            logger.error(f"数据持久化异常")
            logger.error(f"Error: {e}")
            return False

    @staticmethod
    def erase_data_in_file(filename, event_name, cache) -> bool:
        """
            从文件中擦除数据。

        Args:
            filename (str): 文件名。
            event_name (str): 事件名称。
            cache (dict): 缓存数据的字典。

        Returns:
            bool: 如果成功从文件中擦除数据，则返回 True；否则返回 False。

        Raises:
            Exception: 捕获读写YAML文件时产生的未知异常。
        """
        try:
            if filename and cache:
                for filepath, data in cache.items():
                    if os.path.basename(filepath) == filename:
                        tmp = data["items"]
                        if isinstance(tmp, CommentedSeq):
                            for item in list(tmp):
                                if isinstance(item, CommentedMap):
                                    tmp_map = CommentedMap()
                                    for name, content in item.items():
                                        if name != event_name:
                                            tmp_map[name] = content
                                    tmp.remove(item)
                                    tmp.append(tmp_map)
                        with open(file=filepath, mode="wb") as drf:
                            const.REPLY_YAML.dump(data=data, stream=drf)
                            break

        except Exception as e:
            logger.error(f"数据擦除异常")
            logger.error(f"Error: {e}")
            return False

    @property
    def generic_event_names(self) -> List[str]:
        """
        获取所有默认事件的事件名

        Returns:
            List[str]: 包含所有默认事件的名称列表
        """
        return [
            response.event_name
            for data in self._default_generic_data.values()
            for response in data.items.values()
        ]

    @property
    def custom_event_names(self) -> List[str]:
        """
        获取所有自定义通用事件的事件名

        Returns:
            List[str]: 包含所有自定义通用事件的名称列表
        """
        return [
            response.event_name
            for data in self._custom_generic_data.values()
            for response in data.items.values()
        ]

    @property
    def message_event_names(self) -> List[str]:
        """
        获取所有消息事件的事件名

        Returns:
            List[str]: 包含所有消息事件的名称列表
        """
        return [
            response.event_name
            for data in self._condition_specific_data.values()
            for response in data.items.values()
        ]

    @property
    def all_event_names(self) -> List[str]:
        """
        获取所有事件的事件名

        Returns:
            List[str]: 包含所有事件的事件名称的列表。
        """
        return (
            self.generic_event_names
            + self.message_event_names
            + self.custom_event_names
        )



from dicergirl.utils.dicer import Dicer
from dicergirl.reply.manager import manager
from nonebot.adapters import Event


class Docimasy:
    """
    检定结果解析类
    """

    judge_dict = {
        "critical success": 4,
        "very hard success": 3,
        "hard success": 2,
        "success": 1,
        "fail": 0,
        "fatal fail": -1,
    }

    def __init__(self, detail="", judge: str | int = None):
        self.detail = detail

        if isinstance(judge, int):
            self.judge = judge

        if judge:
            self.judge = self.judge_dict[judge]
        else:
            self.judge = None

    def set_judge(self, judge):
        if isinstance(judge, int):
            self.judge = judge

        self.judge = self.judge_dict[judge]

    def __bool__(self):
        if self.judge > 0:
            return True
        else:
            return False

    def __add__(self, toadd):
        if isinstance(toadd, str):
            toadd = toadd.lstrip("\n")
            if toadd:
                if self.detail:
                    self.detail = self.detail.strip("\n") + "\n" + toadd
                else:
                    self.detail = toadd
        elif isinstance(toadd, int):
            self.judge += toadd
        elif isinstance(toadd, Docimasy):
            self.detail += "\n" + toadd.detail
        else:
            raise NotImplementedError
        return self

    def __int__(self):
        return self.judge

    def __repr__(self):
        return self.detail

    def __str__(self):
        return self.detail


def judger(
    event: Event, dice: Dicer, exp: int, name: str = None, reason: str = None
) -> Docimasy:
    """类 COC 模式技能检定结果"""
    result = dice.roll().calc()
    docimasy = Docimasy()

    if isinstance(exp, int):
        if result == 100:
            judge = "大失败!"
            docimasy.set_judge("fatal fail")
        elif exp < 50 and result > 95:
            judge = "大失败!"
            docimasy.set_judge("fatal fail")
        elif result == 1:
            judge = "大成功!"
            docimasy.set_judge("critical success")
        elif result <= exp // 5:
            judge = "极难成功"
            docimasy.set_judge("very hard success")
        elif result <= exp // 2:
            judge = "困难成功"
            docimasy.set_judge("hard success")
        elif result <= exp:
            judge = "成功"
            docimasy.set_judge("success")
        else:
            judge = "失败"
            docimasy.set_judge("fail")

    if reason and not isinstance(exp, int):
        docimasy += manager.process_generic_event(
            "utils.docimasy.reason",
            event=event,
            SkillName=name,
            Value=exp,
            DiceDescription=dice.description(),
            Docimasy=judge,
        )
    elif not reason and isinstance(exp, int):
        docimasy += manager.process_generic_event(
            "utils.docimasy.skill",
            event=event,
            SkillName=name,
            Value=exp,
            DiceDescription=dice.description(),
            Docimasy=judge,
        )
    else:
        docimasy += manager.process_generic_event(
            "utils.docimasy",
            event=event,
            DiceDiscription=dice.description(),
        )

    return docimasy



from pathlib import Path
from datetime import datetime
from multilogging import multilogger

from nonebot.matcher import Matcher
from nonebot.plugin import on, on_request, on_notice, PluginMetadata
from nonebot.adapters import Bot as Bot
from nonebot.adapters.onebot import V11Bot
from nonebot.adapters.onebot.v11 import (
    MessageEvent,
    GroupMessageEvent,
    Event,
    MessageSegment,
)
from nonebot.adapters.onebot.v11.event import (
    FriendRequestEvent,
    GroupRequestEvent,
    GroupDecreaseNoticeEvent,
)
from nonebot.adapters.onebot.v11.exception import ActionFailed
from nonebot.internal.matcher.matcher import Matcher

from .handlers.on import on_startswith
from .handlers.general import show_handler, set_handler, del_handler, roll, shoot

from .reply.manager import manager

from .utils.operator import botoff, boton, get_name, get_status, init, set_name
from .utils.handlers import get_friend_qids, get_group_id, get_mentions, get_user_card
from .utils.formatters import format_msg, format_str
from .utils.loggers import add_logger, get_loggers, remove_logger, loggers
from .utils.admin import (
    add_super_user,
    get_super_users,
    get_uuid,
    is_super_user,
    make_uuid,
    rm_super_user,
)
from .utils.modes import get_mode, set_mode
from .utils.role import rolekp, roleob
from .utils.version import run_shell_command, get_latest_version
from .utils.plugins import modes
from .utils.parser import CommandParser, Commands, Only, Optional, Required, Positional
from .utils.cards import Cards
from .utils.charactors import Character
from .utils.blacklist import blacklist
from .utils.update import require_update
from .utils.settings import DEBUG, debugon, debugoff

from .plugins.parse import get_plugins
from .plugins.operation import install, remove, upgrade as plgupgrade

from .common.exceptions.pluginerror import (
    PluginExistsError,
    PluginInstallFailedError,
    PluginNotFoundError,
    PluginUninstallFailedError,
)
from .common.messages import help_message
from .common.registers import regist_all
from .common.const import DICERGIRL_LOGS_PATH, VERSION

import logging
import sys
import platform
import psutil
import html
import nonebot
import re
import json
import asyncio


__version__ = VERSION
__plugin_meta__ = PluginMetadata(
    name="DicerGirl",
    description="新一代跨平台开源 TRPG 骰娘框架",
    usage="开箱即用",
    type="application",
    homepage="https://gitee.com/unvisitor/dicer",
    supported_adapters={"~onebot.v11"},
)
__author__ = "苏向夜 <fu050409@163.com>"

logger = multilogger(name="DicerGirl", payload="main")
current_dir = Path(__file__).resolve().parent


try:
    driver = nonebot.get_driver()
    initalized = True
except ValueError:
    initalized = False

if initalized:
    # 开发者指令
    testcommand = on_startswith(".test", priority=1, block=True)
    debugcommand = on_startswith(".debug", priority=1, block=True)

    # 消息监视及日志指令
    messagemonitor = on("message", priority=2, block=False)
    dgmessagemonitor = on("message_sent", priority=2, block=False)
    logcommand = on_startswith(".log", priority=2, block=True)

    # 请求及事件处理
    addrequest = on_request(priority=2, block=True)
    kickedevent = on_notice(priority=2, block=False)

    # 机器人管理指令
    superusercommand = on_startswith((".su", ".sudo"), priority=2, block=True)
    botcommand = on_startswith(".bot", priority=2, block=True)
    dismisscommand = on_startswith(
        (".dismiss", ".exit", ".leave", ".bye"), priority=2, block=True
    )
    registcommand = on_startswith((".regist", ".reg"), priority=2, block=True)
    versioncommand = on_startswith((".version", ".v"), priority=2, block=True)

    # 基本指令
    helpcommand = on_startswith((".help", ".h"), priority=3, block=True)
    rollcommand = on_startswith((".r", ".roll"), priority=4, block=True)
    racommand = on_startswith(".ra", priority=3, block=True)
    rhcommand = on_startswith(".rh", priority=3, block=True)
    rahcommand = on_startswith(".rah", priority=3, block=True)
    delcommand = on_startswith((".del", ".delete"), priority=3, block=True)
    showcommand = on_startswith((".show", ".display"), priority=3, block=True)
    setcommand = on_startswith((".set", ".st", ".s"), priority=4, block=True)
    modecommand = on_startswith((".mode", ".m"), priority=3, block=True)
    shootcommand = on_startswith((".sht", ".shoot"), priority=3, block=True)
    attackcommand = on_startswith((".at", ".attack"), priority=3, block=True)
    damcommand = on_startswith((".dam", ".damage"), priority=3, block=True)
    encommand = on_startswith((".en", ".encourage"), priority=3, block=True)

    # 身份组指令
    rolecommand = on_startswith(".role", priority=3, block=True)
    rolekpcommand = on_startswith(".kp", priority=3, block=True)
    roleobcommand = on_startswith(".ob", priority=3, block=True)

    # 定时任务
    scheduler = nonebot.require("nonebot_plugin_apscheduler").scheduler

    @driver.on_startup
    async def _() -> None:
        """`Nonebot2`核心加载完成后的初始化方法"""
        global DEBUG
        logger.info(f"DicerGirl 版本 {VERSION} 初始化中...")
        if DEBUG:
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(sys.stdout, level="DEBUG")
            logger.info("DEBUG 模式已启动.")
        regist_all()
        init()
        logger.success("DicerGirl 初始化完毕.")

    @addrequest.handle()
    async def friendaddapproval(bot: V11Bot, event: FriendRequestEvent):
        if event.user_id in await get_friend_qids(bot):
            return

        if event.get_user_id() in blacklist.get_blacklist():
            try:
                await event.reject(bot)
            except ActionFailed:
                return

            return await bot.send_private_msg(
                user_id=event.user_id,  # TODO
                message=manager.process_generic_event(
                    "main.friend.reject", event=event
                ),
            )

        try:
            await event.approve(bot)
        except ActionFailed:
            return

        super_users = get_super_users()
        for superuser in super_users:
            try:
                await bot.send_private_msg(
                    user_id=superuser,
                    message=manager.process_generic_event(
                        "main.friend.approve", event=event
                    ),
                )
            except ActionFailed:
                pass

        await asyncio.sleep(5.5)

        return await bot.send_private_msg(
            user_id=event.user_id,
            message=manager.process_generic_event("main.friend.new", event=event),
        )

    @addrequest.handle()
    async def groupaddapproval(bot: V11Bot, event: GroupRequestEvent):
        if event.sub_type == "add":
            return

        if (
            event.get_user_id() in blacklist.get_blacklist()
            or event.group_id in blacklist.get_group_blacklist()
        ):
            try:
                await event.reject(bot)
            except ActionFailed:
                await bot.set_group_leave(group_id=event.group_id)

            return await bot.send_private_msg(
                user_id=event.user_id,
                message=manager.process_generic_event("GroupForbidden", event=event),
            )

        try:
            await event.approve(bot)
        except ActionFailed:
            return

        super_users = get_super_users()
        for superuser in super_users:
            await bot.send_private_msg(
                user_id=superuser,
                message=manager.process_generic_event("GroupApproval", event=event),
            )

        if not super_users:
            await bot.send_private_msg(
                user_id=event.self_id,
                message=manager.process_generic_event("GroupApproval", event=event),
            )

    @kickedevent.handle()
    async def onkickhandler(bot: V11Bot, event: GroupDecreaseNoticeEvent):
        if not event.is_tome():
            logger.info(f"用户[{event.user_id}]被移出群聊.")
            return

        if (
            not (event.group_id or event.operator_id)
            or event.user_id == event.operator_id
        ):
            return

        if event.group_id:
            blacklist.add_group_blacklist(str(event.group_id))

        if event.operator_id:
            blacklist.add_blacklist(str(event.operator_id))

        try:
            await bot.call_api("delete_friend", user_id=event.operator_id)
        except ActionFailed:
            return

        super_users = get_super_users()
        for superuser in super_users:
            await bot.send_private_msg(
                user_id=superuser,
                message=manager.process_generic_event(
                    "BlacklistAdded", event=event, UserID=event.operator_id
                ),
            )

        if not super_users:
            await bot.send_private_msg(
                user_id=event.self_id,
                message=manager.process_generic_event(
                    "BlacklistAdded", event=event, UserID=event.operator_id
                ),
            )

    @testcommand.handle()
    async def testhandler(bot: V11Bot, matcher: Matcher, event: MessageEvent):
        """测试指令"""
        args = format_msg(event.get_message(), begin=".test")
        cp = CommandParser(
            Commands(
                [
                    Only("all"),
                    Only("split"),
                    Only("invite"),
                    Only("at"),
                    Only("userid"),
                    Optional("run", str),
                ]
            ),
            args=args,
            auto=True,
        )

        if not is_super_user(event):
            return await matcher.send(
                manager.process_generic_event("PermissionDenied", event=event)
            )

        reply = ""
        if cp.nothing or cp.results["all"]:
            import json

            reply += f"收到消息: {event.get_message()}\n"
            reply += f"消息来源: {event.group_id} - {event.get_user_id()}\n"
            reply += f"消息拆析: {args}\n"
            reply += f"指令拆析: {cp.results}\n"
            reply += f"消息原始内容: {event.get_plaintext()}\n"
            reply += f"消息json数据: {event.json()}\n"
            reply += f"消息发送者json信息: {json.loads(event.json())['sender']}\n"
            reply += f"发送者昵称: {json.loads(event.json())['sender']['nickname']}"
            return await matcher.send(reply)
        else:
            cp = cp.results

        if cp["invite"]:
            await matcher.send(
                "http://qm.qq.com/cgi-bin/qm/qr?_wv=1027&k=hvaf8JGmEXA3N9r4SGgpghDti31aW1bR&authKey=%2Bux%2BedOIguriMYBMGe40coeOT7mx%2B99%2FVMbK0MvE2w1AsVQLLK%2B0hBO6vVB%2Bmlws&noverify=0&group_code=770386358"
            )
            return await bot.send_private_msg(
                user_id=event.user_id,
                message="""[CQ:json,data={"app":"com.tencent.qun.invite"&#44;"config":{"autosize":0&#44;"ctime":1696238280&#44;"round":1&#44;"token":"c05ba4b99859234b1f2b225b74740808"&#44;"type":"normal"}&#44;"meta":{"news":{"desc":"邀请你加入群聊“DicerGirl 公测群 | 未知访客”，进入可查看详情。"&#44;"jumpUrl":"mqqapi://group/invite_join?src_type=internal&amp;version=1&amp;groupcode=770386358&amp;msgseq=1696238280521100&amp;groupname=DicerGirl+%e5%85%ac%e6%b5%8b%e7%be%a4+%7c+%e6%9c%aa%e7%9f%a5%e8%ae%bf%e5%ae%a2&amp;senderuin=1264983312&amp;receiveruin=1557089913"&#44;"preview":"https://p.qlogo.cn/gh/770386358/770386358/?t=1696238280"&#44;"tag":"邀请加群"&#44;"tagIcon":"https://downv6.qq.com/innovate/group_ark_icon.png"&#44;"title":"邀请你加入群聊"}}&#44;"prompt":"邀请加群"&#44;"ver":"1.0.0.31"&#44;"view":"news"}]""",
                auto_escape=False,
            )

        if cp["at"]:
            return await matcher.send(MessageSegment.at(user_id=event.user_id))

        if cp["userid"]:
            return await matcher.send(
                manager.process_generic_event("main.test.userid", event=event)
            )

        if cp["run"]:
            try:
                result = eval(cp["run"])
            except Exception as e:
                result = f"出现异常：{e}"

            return await matcher.send(result)

    @debugcommand.handle()
    async def debughandler(matcher: Matcher, event: MessageEvent):
        """漏洞检测指令"""
        args = format_msg(event.get_message(), begin=".debug")
        if not is_super_user(event):
            await matcher.send(
                manager.process_generic_event("PermissionDenied", event=event)
            )
            return

        if args:
            logger.debug(args)
            if args[0] == "off":
                debugoff()
                logging.getLogger().setLevel(logging.INFO)
                logger.remove()
                logger.add(sys.stdout, level="INFO")
                logger.info("输出等级设置为 INFO.")
                return await matcher.send(
                    manager.process_generic_event("DebugOff", event=event)
                )
        else:
            debugon()
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(sys.stdout, level="INFO")
            logger.info("输出等级设置为 DEBUG.")
            return await matcher.send(
                manager.process_generic_event("DebugOn", event=event)
            )

        if args[0] == "on":
            debugon()
            logging.getLogger().setLevel(logging.DEBUG)
            logger.remove()
            logger.add(sys.stdout, level="INFO")
            logger.info("输出等级设置为 DEBUG.")
            return await matcher.send(
                manager.process_generic_event("DebugOn", event=event)
            )
        else:
            return await matcher.send("错误, 我无法解析你的指令.")

    @superusercommand.handle()
    async def superuser_handler(matcher: Matcher, event: MessageEvent):
        """超级用户管理指令"""
        args = format_str(event.get_message(), begin=(".su", ".sudo"))
        arg = list(filter(None, args.split(" ")))

        if len(arg) >= 1:
            if arg[0].lower() == "exit":
                if not rm_super_user(event):
                    return await matcher.send(
                        manager.process_generic_event("NotManagerYet", event=event)
                    )

                return await matcher.send(
                    manager.process_generic_event("ManagerExit", event=event)
                )

        if is_super_user(event):
            return await matcher.send(
                manager.process_generic_event("AlreadyManager", event=event)
            )

        if not args:
            logger.critical(f"超级令牌: {make_uuid()}")
            await matcher.send(
                manager.process_generic_event("AuthenticateStarted", event=event)
            )
        else:
            if not args == get_uuid():
                await matcher.send(
                    manager.process_generic_event("AuthenticateFailed", event=event)
                )
            else:
                add_super_user(event)
                await matcher.send(
                    manager.process_generic_event("AuthenticateSuccess", event=event)
                )

    @botcommand.handle()
    async def bothandler(
        bot: V11Bot, matcher: Matcher, event: MessageEvent, args: list = []
    ):
        """机器人管理指令"""
        if get_mentions(event) and not event.to_me:
            return

        args = (
            format_msg(event.get_message(), begin=".bot", zh_en=True)
            if not args
            else args
        )
        commands = CommandParser(
            Commands(
                [
                    Only(("version", "v", "bot", "版本")),
                    Only(("exit", "bye", "leave", "离开")),
                    Only(("on", "run", "start", "启动")),
                    Only(("off", "down", "shutdown", "关闭")),
                    Only(("upgrade", "up", "更新")),
                    Optional(("downgrade", "降级"), str),
                    Optional(("name", "命名"), str),
                    Only(("status", "状态")),
                    Optional(("plgup", "pluginup", "升级"), str),
                    Optional(("install", "add", "安装"), str),
                    Optional(("remove", "del", "rm", "删除", "卸载"), str),
                    Only(("mode", "list", "已安装")),
                    Only(("store", "plugins", "商店")),
                    Optional(("search", "搜索"), str),
                    Positional("first", str),
                ]
            ),
            args=args,
            auto=True,
        )

        if commands.nothing or commands.results["version"]:
            return await versionhandler(matcher=matcher, event=event)
        else:
            commands = commands.results

        if commands["exit"]:
            logger.info(f"{get_name()}退出群聊: {event.group_id}")
            await matcher.send(
                manager.process_generic_event("GroupLeaveSet", event=event)
            )
            return await bot.set_group_leave(group_id=event.group_id)

        if commands["on"]:
            boton(event)
            return await matcher.send(
                manager.process_generic_event("BotOn", event=event)
            )

        if commands["off"]:
            botoff(event)
            return await matcher.send(
                manager.process_generic_event("BotOff", event=event)
            )

        if commands["status"]:
            try:
                system = platform.freedesktop_os_release()["PRETTY_NAME"]
            except:
                system = platform.platform()

            memi = psutil.Process().memory_info()
            rss = memi.rss / 1024 / 1024
            total = psutil.virtual_memory().total / 1024 / 1024

            reply = f"DicerGirl {VERSION}, {'正常运行' if get_status(event) else '指令限制'}\n"
            reply += f"骰娘尊名: {get_name()}\n"
            reply += f"操作系统: {system}\n"
            reply += f"CPU 核心: {psutil.cpu_count()} 核心\n"
            reply += f"Python 版本: {platform.python_version()}\n"
            reply += "系统内存占用: %.2fMB/%.2fMB\n" % (rss, total)
            reply += f"漏洞检测模式: {'on' if DEBUG else 'off'}"
            return await matcher.send(reply)

        if commands["upgrade"]:
            await matcher.send("检查版本更新中...")
            newest_version = await get_latest_version("dicergirl")

            if require_update(VERSION, newest_version):
                if {"a", "b"} & set(newest_version) and commands["first"] in (
                    "force",
                    "强制",
                ):
                    return await matcher.send(
                        f"发现新版本 dicergirl {newest_version}, 然而该版本为{'公测' if 'b' in newest_version else '内测'}版本.\n使用`.bot upgrade force`强制更新到该版本."
                    )

                await matcher.send(f"发现新版本 dicergirl {newest_version}, 开始更新...")
                upgrade = await run_shell_command(
                    f'"{sys.executable}" -m pip install dicergirl -i https://pypi.org/simple --upgrade'
                )

                if upgrade["returncode"] != 0:
                    logger.error(upgrade["stderr"])
                    return await matcher.send("更新失败! 请查看终端输出以获取错误信息, 或者你可以再次尝试.")

                return await matcher.send(f"{get_name()}骰娘已更新为版本 {newest_version}.")

            return await matcher.send(f"我已经是最新版本的{get_name()}了!")

        if commands["downgrade"]:
            if not is_super_user(event):
                return await matcher.send("你没有管理员权限, 请先执行`.su`开启权限鉴定.")

            await matcher.send("警告! 执行降级之后可能导致无法再次自动升级!")
            await run_shell_command(
                f"\"{sys.executable}\" -m pip install dicergirl=={commands['downgrade']} -i https://pypi.org/simple"
            )
            await matcher.send(f"{get_name()}已降级到 {commands['downgrade']}.")
            return

        if commands["name"]:
            if not is_super_user(event):
                return await matcher.send("你没有管理员权限, 请先执行`.su`开启权限鉴定.")

            sn = set_name(commands["name"])

            if isinstance(sn, bool):
                await bot.call_api(
                    "set_qq_profile",
                    **{
                        "nickname": commands["name"],
                    },
                )
                await matcher.send(
                    manager.process_generic_event(
                        "NameSet", event=event, NewName=commands["name"]
                    )
                )
            elif isinstance(sn, str):
                await matcher.send(sn)

            return

        if commands["mode"]:
            reply = "当前已正确安装的跑团插件:\n"
            for plugin in modes.keys():
                reply += f"{plugin.upper()} 模式:\n"
                if hasattr(modes[plugin], "__description__"):
                    reply += f"\t简介: {modes[plugin].__description__}\n"

                reply += f"\t切换: .mode {plugin}\n"

            reply.strip("\n")
            reply += f"当前的跑团模式为 {get_mode(event).upper()}."
            return await matcher.send(reply)

        if commands["store"]:
            official, community = await get_plugins()
            reply = "从商店搜索到以下插件:\n"
            reply += "Unvisitor 官方插件:\n"

            i = 1
            for name, detail in official.items():
                reply += f"  {i}. {detail['name']}[安装: .bot install {name}]\n"
                i += 1

            reply += "社区提供插件:\n"
            i = 1
            for name, detail in community.items():
                reply += f"  {i}. {detail['name']}[安装: .bot install {name}]\n"

            reply.rstrip("\n")
            return await matcher.send(reply)

        if commands["install"]:
            if commands["install"] in modes.keys():
                return await matcher.send(f"模块 {commands['install']} 已经安装了!")

            ins = await install(commands["install"])

            if ins is PluginNotFoundError:
                return await matcher.send(f"包 {commands['install']} 不存在.")
            elif ins is PluginInstallFailedError:
                return await matcher.send(f"包 {commands['install']} 安装失败.")
            elif ins == True:
                return await matcher.send(f"模块 {commands['install']} 安装完毕.")

            return

        if commands["remove"]:
            if not is_super_user(event):
                return await matcher.send("你没有管理员权限, 请先执行`.su`开启权限鉴定.")

            uns = await remove(commands["remove"])

            if uns is PluginUninstallFailedError:
                return await matcher.send(
                    manager.process_generic_event("UninstallFailed", event=event)
                )

            return await matcher.send(f"模块 {commands['remove']} 卸载完毕.")

        if commands["plgup"]:
            await matcher.send(f"检查插件 {commands['plgup']} 更新中...")
            if commands["plgup"] not in modes.keys():
                plg_version = None
            else:
                plg_version = modes[commands["plgup"]].__version__

            latest_version = await get_latest_version(
                f"dicergirl-plugin-{commands['plgup']}"
            )

            if latest_version == plg_version:
                return await matcher.send(
                    f"插件[{commands['plgup']}]已经是最新版本了.\n`.bot remove {commands['plgup']}`并`.bot install {commands['plgup']}`可以重新安装插件.\n如果遇到问题, 使用`.help 支持`获得开发者联系方式."
                )
            else:
                await matcher.send(
                    f"发现插件[{commands['plgup']}]新版本[{latest_version}], 开始更新."
                )

            up = await plgupgrade(commands["plgup"])

            if up is PluginNotFoundError:
                return await matcher.send(f"包[{commands['plgup']}]似乎不存在?")
            elif up is PluginInstallFailedError:
                return await matcher.send(f"包[{commands['plgup']}]更新失败了.")
            elif up == True:
                return await matcher.send(f"插件[{commands['plgup']}]更新完毕.")

        return await matcher.send("未知的指令, 使用`.help bot`获得机器人管理指令使用帮助.")

    @dismisscommand.handle()
    async def dismisshandler(bot: V11Bot, matcher: Matcher, event: Event):
        await bothandler(bot=bot, matcher=matcher, event=event, args=["exit"])

    @logcommand.handle()
    async def loghandler(bot: V11Bot, matcher: Matcher, event: Event):
        """日志管理指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".log")
        commands = CommandParser(
            Commands(
                [
                    Only("show"),
                    Only(("add", "new")),
                    Optional("name", str),
                    Optional("stop", int),
                    Optional("start", int),
                    Optional(("remove", "rm"), int),
                    Optional(("download", "load"), int),
                ]
            ),
            args=args,
            auto=True,
        ).results

        if commands["show"]:
            gl = get_loggers(event)
            if len(gl) == 0:
                return await matcher.send("暂无存储的日志.")

            if not get_group_id(event) in loggers.keys():
                running = []
            else:
                running = loggers[get_group_id(event)].keys()

            reply = "该群聊所有日志:\n"
            for l in gl:
                index = gl.index(l)
                reply += f"序列 {index} : {l} : {'记录中' if index in running else '已关闭'}\n"
            reply.strip("\n")

            return await matcher.send(reply)

        if commands["download"] or commands["download"] == 0:
            gl = get_loggers(event)
            if commands["download"] > len(gl) - 1:
                return await matcher.send(f"目标日志序列 {commands['download']} 不存在.")

            path = Path(gl[commands["download"]])
            return await bot.call_api(
                "upload_group_file",
                **{
                    "group_id": get_group_id(event),
                    "file": str(path),
                    "name": path.name,
                },
            )

        if commands["add"]:
            if commands["name"]:
                logname = str(DICERGIRL_LOGS_PATH / (commands["name"] + ".trpg.log"))
            else:
                logname = str(
                    DICERGIRL_LOGS_PATH
                    / (datetime.now().strftime("%Y-%m-%d-%H-%M-%S") + ".trpg.log")
                )

            new_logger = multilogger(name="DG Msg Logger", payload="TRPG")
            new_logger.remove()
            new_logger.add(logname, encoding="utf-8", format="{message}", level="INFO")
            if not get_group_id(event) in loggers.keys():
                loggers[get_group_id(event)] = {}

            index = len(get_loggers(event))
            loggers[get_group_id(event)][index] = [new_logger, logname]

            if not add_logger(event, logname):
                raise IOError("无法新增日志.")

            return await matcher.send(f"新增日志序列: {index}\n日志文件: {logname}")

        if commands["stop"] or commands["stop"] == 0:
            if not get_group_id(event) in loggers.keys():
                await matcher.send("该群组无正在运行中的日志.")
                loggers[get_group_id(event)] = {}
                return

            if not commands["stop"] in loggers[get_group_id(event)]:
                await matcher.send("目标日志不存在.")
                return

            loggers[get_group_id(event)][commands["stop"]][0].remove()
            loggers[get_group_id(event)].pop(commands["stop"])
            await matcher.send(f"日志序列 {commands['stop']} 已停止记录.")
            return

        if commands["start"] or commands["start"] == 0:
            gl = get_loggers(event)
            if commands["start"] > len(gl) - 1:
                await matcher.send(f"目标日志序列 {commands['start']} 不存在.")
                return

            logname = gl[commands["start"]]
            new_logger = multilogger(name="DG Msg Logger", payload="TRPG")
            new_logger.remove()
            new_logger.add(logname, encoding="utf-8", format="{message}", level="INFO")
            if not get_group_id(event) in loggers.keys():
                loggers[get_group_id(event)] = {}

            loggers[get_group_id(event)][commands["start"]] = [new_logger, logname]
            await matcher.send(f"日志序列 {commands['start']} 重新启动.")
            return

        if commands["remove"] or commands["remove"] == 0:
            gl = get_loggers(event)
            if not gl:
                await matcher.send("该群组从未设置过日志.")
                loggers[get_group_id(event)] = {}
                return

            if commands["remove"] > len(gl) - 1:
                return await matcher.send(f"目标日志序列 {commands['remove']} 不存在.")

            index = len(gl)
            if get_group_id(event) in loggers.keys():
                if commands["remove"] in loggers[get_group_id(event)].keys():
                    await matcher.send(f"目标日志序列 {commands['remove']} 正在运行, 开始中止...")
                    loggers[get_group_id(event)][commands["remove"]][0].remove()
                    loggers[get_group_id(event)].pop(commands["remove"])

            Path(gl[commands["remove"]]).unlink()
            remove_logger(event, commands["remove"])
            return await matcher.send(f"日志序列 {commands['remove']} 已删除.")

        return await matcher.send("骰娘日志管理系统, 使用`.help log`指令详细信息.")

    def trpg_log(event) -> None:
        """外置的日志记录方法"""
        if not get_group_id(event) in loggers.keys():
            return

        for log in loggers[get_group_id(event)].keys():
            if isinstance(event, GroupMessageEvent):
                raw_json = json.loads(event.json())
                if raw_json["sender"]["card"]:
                    if raw_json["sender"]["card"].lower() == "ob":
                        role_or_name = f"[旁观者 - {raw_json['sender']['nickname']}]"
                    elif raw_json["sender"]["card"].lower() == "kp":
                        role_or_name = f"[主持人 - {raw_json['sender']['nickname']}]"
                    else:
                        role_or_name = f"[{raw_json['sender']['card']}]"
                elif raw_json["sender"]["nickname"]:
                    role_or_name = f"[未知访客 - {raw_json['sender']['nickname']}]"
                else:
                    role_or_name = f"[未知访客 - {str(event.get_user_id())}]"

                message = role_or_name + ": " + html.unescape(str(event.get_message()))
            elif isinstance(event, Event):
                message = f"[{get_name()}]: " + html.unescape(event.message)

            loggers[get_group_id(event)][log][0].info(message)

    @dgmessagemonitor.handle()
    @messagemonitor.handle()
    def loggerhandler(event: Event):
        """消息记录日志指令"""
        return trpg_log(event)

    @showcommand.handle()
    async def showhandler(
        matcher: Matcher, event: MessageEvent, args: list = None
    ):
        """角色卡展示指令"""
        if not get_status(event) and not event.to_me:
            return

        if not isinstance(args, list):
            args = format_msg(event.get_message(), begin=(".show", ".display"))
        else:
            args = args

        at = get_mentions(event)

        mode = get_mode(event)
        if mode in modes:
            try:
                sh = show_handler(event, args, at, mode=mode)
            except Exception as error:
                logger.exception(error)
                sh = [manager.process_generic_event("CommandFailed", event=event)]
        else:
            return await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

        for msg in sh:
            await matcher.send(str(msg))

        return

    @setcommand.handle()
    async def sethandler(bot: V11Bot, matcher: Matcher, event: MessageEvent):
        """角色卡设置指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".set", ".st", ".s"), zh_en=False)
        at = get_mentions(event)
        commands = CommandParser(
            Commands(
                [
                    Only("show"),
                    Only("del"),
                    Only("clear"),
                    Only("init"),
                    Optional(("mode", "temp"), str),
                    Optional(("name", "n"), str),
                ]
            ),
            args,
            auto=True,
        ).results

        if at and not get_user_card(event).upper() == "KP" and not event.to_me:
            await matcher.send(manager.process_generic_event("SetPermissionDenied"))
            return
        if len(at) == 1:
            qid = at[0]
        else:
            qid = ""

        mode = get_mode(event)

        if commands["name"]:
            cards: Cards = modes[mode].__cards__
            if not cards.get(event):
                cha: Character = modes[mode].__charactor__()
                if hasattr(cha, "init"):
                    cha.init()

                cards.update(event, cha.__dict__, save=True)

            cha: Character = modes[mode].__charactor__()
            cha.load(cards.get(event, qid=qid))
            setattr(cha, "name", commands["name"])
            await bot.set_group_card(
                group_id=event.group_id, user_id=event.user_id, card=commands["name"]
            )
            return await matcher.send(
                manager.process_generic_event(
                    "main.set.name", event=event, Name=commands["name"]
                )
            )

        if commands["init"]:
            cards: Cards = modes[mode].__cards__
            if not cards.get(event):
                cha: Character = modes[mode].__charactor__()
                if hasattr(cha, "init"):
                    cha.init()

                cards.update(event, cha.__dict__, save=True)

            return await matcher.send(
                manager.process_generic_event("CardInit", event=event)
            )

        if commands["show"]:
            args.remove("show")
            return await showhandler(matcher, event, args=args)

        if commands["del"]:
            args.remove("del")
            return await delhandler(matcher, event, args=args)

        if commands["mode"]:
            if "mode" in args:
                args.remove("mode")
            else:
                args.remove("temp")
            return await modehandler(
                bot=bot, matcher=matcher, event=event, args=[commands["mode"]]
            )

        if mode in modes:
            try:
                if not args:
                    cache = modes[mode].__cache__
                    user_id: int = event.user_id
                    got = cache.get(event, qid=str(user_id))

                    if isinstance(got, dict):
                        if "name" not in got.keys():
                            got = ""

                    name = got["name"] if got else ""
                    await bot.set_group_card(
                        group_id=event.group_id, user_id=user_id, card=name
                    )

                cards: Cards = modes[mode].__cards__
                if not cards.get(event):
                    cha: Character = modes[mode].__charactor__()
                    if hasattr(cha, "init"):
                        cha.init()

                    cards.update(event, cha.__dict__, save=True)

                sh = set_handler(event, args, at, mode=mode)
            except Exception as error:
                logger.exception(error)
                sh = [manager.process_generic_event("CommandFailed", event=event)]
        else:
            return await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

        return await matcher.send(sh)

    @helpcommand.handle()
    async def helphandler(matcher: Matcher, event: MessageEvent):
        """帮助指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".help", ".h"))
        if args:
            arg = args[0]
        else:
            arg = ""

        message = help_message(arg)
        message = re.sub(r"\{name\}", get_name(), message)
        message = re.sub(r"\{version\}", VERSION, message)
        message = re.sub(r"\{py_version\}", platform.python_version(), message)
        message = re.sub(r"\{nonebot_version\}", nonebot.__version__, message)
        return await matcher.send(message)

    @modecommand.handle()
    async def modehandler(
        bot: V11Bot, matcher: Matcher, event: MessageEvent, args: list = []
    ):
        """跑团模式切换指令"""
        if not get_status(event) and not event.to_me:
            return

        args = (
            format_msg(event.get_message(), begin=(".mode", ".m")) if not args else args
        )
        cp = CommandParser(
            Commands(
                [
                    Positional("mode", str),
                ]
            ),
            args=args,
            auto=True,
        )

        if not cp.nothing:
            cp = cp.results
            if cp["mode"] in modes:
                set_mode(event, cp["mode"])

                for user in await bot.get_group_member_list(group_id=event.group_id):
                    if not hasattr(modes[cp["mode"]], "__cards__"):
                        break

                    card: Cards = modes[cp["mode"]].__cards__
                    user_id: int = user["user_id"]
                    got = card.get(event, qid=str(user_id))

                    if isinstance(got, dict):
                        if "name" not in got.keys():
                            continue
                    elif not got:
                        continue

                    name = got["name"] if got else ""
                    await bot.set_group_card(
                        group_id=event.group_id, user_id=user_id, card=name
                    )

                await matcher.send(
                    manager.process_generic_event(
                        "ModeChanged", event=event, Mode=cp["mode"].upper()
                    )
                )
                return True
            else:
                await matcher.send(
                    manager.process_generic_event(
                        "UnknownMode", event=event, Mode=cp["mode"].upper()
                    )
                )
                return True
        else:
            reply = "当前已正确安装的跑团插件:\n"
            for plugin in modes.keys():
                reply += f"{plugin.upper()} 模式: {plugin}.\n"

            reply += f"当前的跑团模式为 {get_mode(event).upper()}."
            await matcher.send(reply)

    @shootcommand.handle()
    async def shoothandler(matcher: Matcher, event: MessageEvent):
        """射击检定指令"""
        if not get_status(event) and not event.to_me:
            return

        await matcher.send(shoot(event))

    @attackcommand.handle()
    async def attackhandler(matcher: Matcher, event: GroupMessageEvent):
        """伤害检定指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_str(event.get_message(), begin=(".at", ".attack"))
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                return await matcher.send(f"跑团模式 {mode.upper()} 未设置标准指令.")

            if not "at" in modes[mode].__commands__.keys():
                return await matcher.send(f"跑团模式 {mode.upper()} 不支持伤害检定指令.")

            handler = modes[mode].__commands__["at"]
            return await matcher.send(handler(event, args))
        else:
            return await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

    @damcommand.handle()
    async def damhandler(matcher: Matcher, event: MessageEvent):
        """承伤检定指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".dam", ".damage"))
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                return await matcher.send(f"跑团模式 {mode.upper()} 未设置标准指令.")

            if not "at" in modes[mode].__commands__.keys():
                return await matcher.send(f"跑团模式 {mode.upper()} 不支持承伤检定指令.")

            handler = modes[mode].__commands__["dam"]
            await matcher.send(handler(event, args))
        else:
            await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

    @encommand.handle()
    async def enhandler(matcher: Matcher, event: MessageEvent):
        """属性或技能激励指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".en")
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                return await matcher.send(f"跑团模式 {mode.upper()} 未设置标准指令.")

            if not "at" in modes[mode].__commands__.keys():
                return await matcher.send(f"跑团模式 {mode.upper()} 不支持激励指令.")

            handler = modes[mode].__commands__["en"]
            await matcher.send(handler(event, args))
        else:
            await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

    @racommand.handle()
    async def rahandler(matcher: Matcher, event: MessageEvent):
        """属性或技能检定指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".ra")
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                return await matcher.send(f"跑团模式 {mode.upper()} 未设置标准指令.")

            if not "ra" in modes[mode].__commands__.keys():
                return await matcher.send(f"跑团模式 {mode.upper()} 不支持技能检定指令.")

            handler = modes[mode].__commands__["ra"]
            replies = handler(event, args)
            if isinstance(replies, list):
                for reply in replies:
                    await matcher.send(reply)

                return

            await matcher.send(replies)
        else:
            await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

    @rhcommand.handle()
    async def rhhandler(bot: Bot, matcher: Matcher, event: MessageEvent):
        """暗骰指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_str(event.get_message(), begin=".rh")
        await matcher.send("暗骰: 命运的骰子在滚动.")
        return await bot.send_private_msg(
            user_id=event.get_user_id(), message=roll(event, args)
        )

    @rahcommand.handle()
    async def rhahandler(bot: Bot, matcher: Matcher, event: GroupMessageEvent):
        """暗骰技能检定指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=".rah")
        mode = get_mode(event)
        if mode in modes:
            if not hasattr(modes[mode], "__commands__"):
                await matcher.send(f"跑团模式 {mode.upper()} 未设置标准指令.")
                return

            if not "ra" in modes[mode].__commands__.keys():
                await matcher.send(f"跑团模式 {mode.upper()} 不支持技能检定指令.")
                return

            await matcher.send("暗骰: 命运的骰子在滚动.")

            handler = modes[mode].__commands__["ra"]
            replies = handler(event, args)
            if isinstance(replies, list):
                for reply in replies:
                    await bot.send_private_msg(
                        user_id=event.get_user_id(), message=reply
                    )
                return

            await bot.send_private_msg(user_id=event.get_user_id(), message=replies)
        else:
            await matcher.send(
                manager.process_generic_event("UnknownMode", event=event)
            )

    @rollcommand.handle()
    async def rollhandler(matcher: Matcher, event: MessageEvent):
        """标准掷骰指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_str(event.get_message(), begin=(".r", ".roll"))
        name = get_user_card(event)
        if not args:
            return await matcher.send(roll(event, "1d100", name=name))

        try:
            await matcher.send(roll(event, args, name=name))
        except Exception as error:
            logger.exception(error)
            await matcher.send(
                "未知错误, 可能是掷骰语法异常.\nBUG提交: https://gitee.com/unvisitor/issues"
            )

    @delcommand.handle()
    async def delhandler(matcher: Matcher, event: MessageEvent, args: list = None):
        """角色卡或角色卡技能删除指令"""
        if not get_status(event) and not event.to_me:
            return

        if not isinstance(args, list):
            args = format_msg(event.get_message(), begin=(".del", ".delete"))
        else:
            args = args

        at = get_mentions(event)

        if at and not is_super_user(event):
            await matcher.send(
                manager.process_generic_event("DeletePermissionDenied", event=event)
            )
            return

        mode = get_mode(event)
        if mode in modes:
            for msg in del_handler(event, args, at, mode=mode):
                await matcher.send(msg)

    @rolecommand.handle()
    async def rolehandler(bot: V11Bot, matcher: Matcher, event: MessageEvent):
        """身份组认证"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".role"))
        cp = CommandParser(
            Commands(
                [
                    Only("kp"),
                    Only("ob"),
                    Only("exit"),
                ]
            ),
            args=args,
            auto=True,
        ).results

        if cp["kp"]:
            return await rolekphandler(bot, matcher, event)

        if cp["ob"]:
            return await roleobhandler(bot, matcher, event)

    @rolekpcommand.handle()
    async def rolekphandler(bot: V11Bot, matcher: Matcher, event: GroupMessageEvent):
        """KP 身份组认证"""
        if not get_status(event) and not event.to_me:
            return

        args = format_msg(event.get_message(), begin=(".kp"))
        cp = CommandParser(
            Commands(
                [
                    Optional(("time", "schedule"), int),
                    Optional(("minute", "min"), int, 0),
                ]
            ),
            args=args,
            auto=True,
        ).results

        if cp["time"]:

            @scheduler.scheduled_job(
                "cron", hour=cp["time"], minute=cp["minute"], id="timer"
            )
            async def _():
                await bot.send_group_msg(
                    group_id=event.group_id, message="抵达 KP 设定的跑团启动时间."
                )

            try:
                scheduler.start()
            except:
                pass

            return await matcher.send(
                f"定时任务: {cp['time']}: {cp['minute'] if len(str(cp['minute'])) > 1 else '0'+str(cp['minute'])}"
            )

        rolekp(event)
        await bot.set_group_card(
            group_id=event.group_id, user_id=event.get_user_id(), card="KP"
        )
        return await matcher.send(manager.process_generic_event("RoleKP", event=event))

    @roleobcommand.handle()
    async def roleobhandler(bot: V11Bot, matcher: Matcher, event: GroupMessageEvent):
        """OB 身份组认证"""
        if not get_status(event) and not event.to_me:
            return

        import json

        roleob(event)

        if json.loads(event.json())["sender"]["card"] == "ob":
            await bot.set_group_card(
                group_id=event.group_id, user_id=event.get_user_id()
            )
            return await matcher.send("取消旁观者 (OB) 身份.")
        else:
            await bot.set_group_card(
                group_id=event.group_id, user_id=event.get_user_id(), card="ob"
            )
            return await matcher.send(
                manager.process_generic_event("RoleOB", event=event)
            )

    @registcommand.handle()
    async def registhandler(matcher: Matcher, event: MessageEvent):
        """消息事件注册指令"""
        if not get_status(event) and not event.to_me:
            return

        args = format_str(
            event.get_message(), begin=(".regist", ".reg"), lower=False
        ).split(" ")
        args = list(filter(None, args))
        cp = CommandParser(
            Commands(
                [
                    Positional("event_name", str),
                    Positional("message", str),
                    Optional(("remove", "rm", "delete", "del"), str),
                    Optional("enable", str),
                    Optional("disable", str),
                ]
            ),
            args=args,
            auto=True,
        ).results

        if cp["remove"]:
            if not manager.remove_event(cp["remove"]):
                return await matcher.send("错误的消息类型, 使用`.help regist`获取帮助信息.")

            return await matcher.send(f"消息事件 {cp['remove']} 已经成功销毁, 该事件将采用默认回复替代.")

        if cp["enable"]:
            if not manager.enable_event(cp["enable"]):
                matcher.send("启用消息事件时出现异常, 请检查事件名是否正确!\n使用`.help regist`获得帮助信息.")
            return await matcher.send(f"消息事件 {cp['enable']} 已启用.")

        if cp["disable"]:
            if not manager.enable_event(cp["disable"]):
                matcher.send("禁用消息事件时出现异常, 请检查事件名是否正确!\n使用`.help regist`获得帮助信息.")
            return await matcher.send(f"消息事件 {cp['disable']} 已启用.")

        event_name = cp["event_name"]
        message = cp["message"]
        if not event_name or not message:
            return await matcher.send("消息事件注册参数不全, 使用`.help regist`获取帮助信息.")

        message = html.unescape(message)
        manager.register_event(event_name, message.replace("\\n", "\n"), is_custom=True)
        return await matcher.send(f"消息事件 {event_name} 已被更改为 {message}.")

    @versioncommand.handle()
    async def versionhandler(matcher: Matcher, event: GroupMessageEvent):
        """骰娘版本及开源声明指令"""
        return await matcher.send(
            f"Unvisitor DicerGirl 版本 {VERSION} [Python {platform.python_version()} For Nonebot2 {nonebot.__version__}]\n此项目以Apache-2.0协议开源.\nThis project is open source under the Apache-2.0 license.\n欢迎使用 DicerGirl, 使用`.help 指令`查看指令帮助."
        )

else:
    logger.warning("Nonebot2 初始化失败, DicerGirl 无法启动!")



from typing import Dict, List
from ..reply.manager import manager

import Levenshtein

manager.register_event(
    "HelpNotFoundRelated", "{BotName}没有找到相关帮助, 你是否是指: {MostRelated}?"
)
manager.register_event(
    "HelpNotFoundSomeRelated",
    "{BotName}没有找到相关帮助, 你是否是指:\n{MostRelated}\n使用.help [以上参数]来获得相关帮助内容.",
)


def similar(str1, str2):
    distance = Levenshtein.distance(str1, str2)
    return 1 - (distance / max(len(str1), len(str2)))


class Messages:
    """骰娘帮助信息"""

    keys: Dict[str, List[str]] = {
        "帮助": ["main", "", None, "帮助"],
        "指令": ["commands", "command", "cmd", "指令"],
        "管理": ["admin", "管理"],
        "支持": ["supports", "support", "spt", "支持"],
        "管理员鉴权": ["sudo", "su", "管理员鉴权", "鉴权"],
        "模式": ["mode", "m", "模式"],
        "机器人管理": ["bot", "manage", "机器人管理"],
        "录卡": ["set", "st", "录卡"],
        "展示": ["show", "展示"],
        "掷骰": ["roll", "r", "rd", "掷骰"],
        "检定": ["ra", "技能检定", "检定"],
        "承伤": ["dam", "damage", "承伤检定", "承伤检定"],
        "伤害": ["at", "attack", "伤害检定", "伤害"],
        "激励": ["en", "激励检定", "激励检定"],
        "删除": ["delete", "del", "remove", "rm", "删除"],
        "日志": ["log", "logger", "日志管理", "日志系统", "日志"],
        "消息": ["regist", "reg", "消息事件", "回复", "消息"],
    }
    main = """Unvisitor DicerGirl 版本 {version} [Python {py_version} For Nonebot2 {nonebot_version}]
.help (.h)  展示此帮助信息
.help 指令  查看指令帮助
.help 管理  查看骰娘管理指令
.help 支持  获取开发者支持
使用`.help [模式名称]`以获取该跑团模式的帮助.
此项目以 Apache-2.0 协议开源.
Copyright 2011-2023 Unknown Visitor, org."""
    commands = """Unvisitor DicerGirl 版本 {version}
角色卡设定或模式切换后骰娘会自动修改群名片, 在保存新的人物卡时同样会更改群名片.
所有指令允许大小写混用.
.mode (.m)  切换和查询跑团模式
.set (.st)  角色卡设定
.show (.st show)  角色卡查询
.roll (.r)  掷骰指令
.ra  属性或技能检定
.at (.attack)  角色伤害检定
.dam (.damage)  角色承伤检定
.en (.encourage)  成长检定
.del (.st del)  删除数据
.kp  进入主持人席位
.ob  进入旁观者席位
.log  日志管理
.reg (.regist)  消息事件注册
使用`.help [指令名]`获取该指令的详细信息"""
    admin = """Unvisitor DicerGirl 版本 {version}
骰娘管理指令, 如果你是骰主或开发者, 请注意 DicerGirl 的管理员与 Nonebot2 不同.
.su/.sudo  进行骰娘管理员鉴权
.bot  机器人管理
使用`.help [指令名]`获取该指令的详细信息"""
    supports = """Unvisitor DicerGirl 版本 {version}
开发团队: 未知访客黑客联盟
团队官网: https://www.unvisitor.site/
DicerGirl 主页: https://dicer.unvisitor.site/
DicerGirl 项目主页:
  1. Gitee: https://gitee.com/unvisitor/dicer
  2. Github: https://github.com/unvisitor/dicer
BUG 提交: https://gitee.com/unvisitor/dicer/issues
功能建议: https://gitee.com/unvisitor/dicer/issues
公测 QQ 群: 770386358
项目负责人: 1264983312"""
    sudo = """用法：.su [鉴权令牌]
描述：
    进行管理员鉴权, 获得骰娘的管理员权限。
示例:
    .su d32ab3...
注意：
    - 鉴权令牌会在执行无参数的`.su`指令后, 在`Nonebot2`的控制终端输出, 输出模式为`CRITICAL`.
    - 需要注意此权限管理系统与`Nonebot2`的`SUPERUSER`不同."""
    bot = """用法：.bot <指令> [参数]
描述：
    执行与管理插件和机器人设置相关的各种任务。
指令：
    version (v, bot, 版本)          显示机器人版本
    exit (bye, leave, 离开)         退出机器人
    on (run, start, 启动)           启动机器人
    off (down, shutdown, 关闭)      关闭机器人
    upgrade (up, 更新)              升级机器人
    downgrade (降级)                降级机器人
    name (命名) <名称>              设置或显示机器人名称
    status (状态)                   显示机器人当前状态
    plgup (pluginup, 升级) <名称>    升级特定插件
    install (add, 安装) <名称>      安装插件
    remove (del, rm, 删除, 卸载) <名称>    删除插件
    mode (list, 已安装)             列出已安装的插件
    store (plugins, 商店)           显示商店中可用的插件
    search (搜索) <名称>            在商店中搜索插件
示例：
    .bot version
    .bot install 插件名称
    .bot remove 插件名称"""
    mode = """用法：.mode (.m) [模式名称]
描述：
    切换跑团模式。
示例：
    .mode coc  切换到 COC 跑团模式
注意：
    - 如果骰娘管理员加入了第三方跑团插件, `mode`参数应该设置为该插件中`__init__.py`的`__name__`参数, 不区分大小写."""
    set = """用法：.set (.st) <指令> [属性名称] [属性值] ... [CQ:at]
描述：
    设置角色卡信息。
指令：
    show (=.show)    显示角色卡信息
    del (=.del)     删除角色卡
    clear    清空所有角色卡信息
示例：
    .set 毁灭人类 99 打爆地球 99
    .set 幸运 +10
    .set show
    .set del 毁灭人类
    .set clear
注意：
    - 在保存人物卡之前请先确保当前模式与车卡模式相同.
    - 部分跑团模式(如SCP)中不支持设置属性, 同时设置非自定义技能也是不推荐的, 在设置前, 建议先询问主持人的意见.
    - 在群聊中输入单独的`.set`指令, 欧若可将自动读取最近一次车卡(即人物卡作成)指令的结果进行保存.
    - 当属性或技能名称均为中文或均为英文时, 指令是强空格需求的, 或者你也可以使用引号规避指令解析问题.
    - 录卡一般常见于类似 COC 跑团和 DND 跑团等跑团模式, 部分模式(如SCP模式)是不需要的, 在建卡之前, 请先询问主持人是否需要进行录卡.
    - `[CQ:at]`指在群聊中`@`一个玩家, `.set`指令将会指向该玩家, 该参数是可选的."""
    show = """.show (.st show) [选项] [CQ:at]
描述：
    展示人物卡
指令：
    [键值]   该模式下存在的可查看键值
    skill   查看自身人物卡技能
    all     查询所有存储的人物卡
注意：
    - 部分键值可能并不在其它模式中支持, 如果管理员加入了第三方插件, 准允的可选键值请询问主持人、骰娘管理员或插件开发者.
    - 官方插件的键值可在插件主页查看
    - `[CQ:at]`指在群聊中`@`一个玩家, `.show`指令将会指向该玩家, 该参数是可选的."""
    roll = """.roll (.r) [掷骰表达式]
描述：
    标准掷骰指令。
指令：
    [掷骰表达式]    掷骰表达式
示例：
    .r 1d100    掷骰1d100
    .r 1d20+1d6-3d10    复杂掷骰计算
    .r (1d6+2d10)/2-3d6*2     复合掷骰运算
备注：
    - 掷骰表达式兼容常用的OneDice标准.
    - 空的掷骰表达式默认为`1d100`.
"""
    ra = """.ra <属性/技能名> [检定值]
描述：
    基础属性或技能检定。
指令：
    <属性/技能名>   人物卡中包含的基础属性或技能名称
    [检定值]      一般是指定属性/技能值或检定难度
示例：
    .ra 命运    快速检定`命运`属性
    .ra 毁灭人类 99     检定技能值为99的毁灭人类技能"""
    dam = """.dam (.damage) [选项] [掷骰表达式]
描述：
    角色承伤检定。
指令：
    [掷骰表达式]    掷骰表达式
    check   检定人物当前生命状态
示例：
    .dam check
    .dam 1d6+2  人物受到`1d6+2`掷骰结果的伤害
    .dam 6  人物受到 6 点伤害
注意：
    - 无参数的`.dam`指令会进行该模式默认的伤害承受检定."""
    at = """.at (.attack) [掷骰表达式] [参数]
描述：
    角色伤害检定。
指令：
    [掷骰表达式]    掷骰表达式
    [参数]      一般为检定工具伤害
示例：
    .at 1d6     人物造成`1d6`掷骰结果的伤害
    .at 燃烧瓶      使用燃烧瓶进行伤害检定
注意：
    - 无参数的`.at`指令会进行该模式默认的近战伤害检定.
    - 部分模式可能不支持掷骰表达式或语法."""
    en = """.en (.encourage) <属性/技能名> [激励点]
描述：
    角色(临时)属性激励。
指令：
    <属性/技能名>   角色属性或技能名
    [激励点]        激励点
示例：
    .en 强度 2  激励属性`强度`2 点
    .en 毁灭人类    激励技能`毁灭人类`
注意：
    - 部分模式可能不需要给入激励点."""
    delele = """.del (.delete) [选项] [参数]
描述：
    删除数据。
指令：
    cache   删除暂存数据
    card    删除使用中的人物卡(谨慎使用)
    [技能]  删除指定的自定义技能
注意：
    - 删除自定义技能时, 支持多个参数, 可以一次指定多个技能名."""
    log = """用法：.log <指令> [选项]
描述：
    执行与管理日志相关的各种任务。
指令：
    show          显示所有日志
    add (new) [名称]   添加一个新日志，可选择提供一个名称
    stop <ID>     停止特定 ID 的日志记录
    start <ID>    启动特定 ID 的日志记录
    remove (rm) <ID>  删除特定 ID 的日志记录
    download (load) <ID> 下载特定 ID 的日志记录
示例：
    .log show
    .log add 日志名称
    .log stop 日志ID
注意：
    - 使用 'add' 指令而不带参数以创建一个未命名的日志。"""
    regist = """用法：.regist (.reg) [事件名] [消息内容] [指令]
描述：
    用户自定义回复文本。
指令：
    remove (rm, delete, del) [事件名]   注销自定义消息事件
    enable [事件名]      启用被禁用的消息事件
    disable [事件名]     禁用自定义消息事件
示例：
    .regist ModeChanged 已切换跑团模式为{Mode}.     将默认`ModeChanged`事件回复内容修改
    .regist remove ModeChanged      返回默认消息内容"""

    def get(self, key) -> None | str:
        for _, alias in self.keys.items():
            if key in alias:
                return self.__getattribute__(alias[0])

        return


messages = Messages()


def help_message(args: str) -> str:
    """`.help`指令后端方法"""
    similarcmd = {}
    got = messages.get(args)
    if got:
        return got

    for key, alias in messages.keys.items():
        relation = []
        for alia in alias:
            if alia:
                relation.append(similar(alia, args))

        similarcmd[key] = max(relation)

    related = []
    most_related = ["", 0]
    for cmd, similarity in similarcmd.items():
        if similarity > 0.8:
            related.append(cmd)

        if most_related[1] < similarity:
            most_related = [cmd, similarity]

    if not related:
        return manager.process_generic_event(
            "HelpNotFoundRelated", MostRelated=most_related[0]
        )

    i = 1
    reply = ""
    for relate in related:
        reply += f"{i}. {relate}\n"
        i += 1

    reply = manager.process_generic_event("HelpNotFoundSomeRelated", MostRelated=reply)
    return reply


def regist(name, message, alias=[]):
    if not alias:
        alias = [
            name,
        ]

    messages.keys[name] = alias
    setattr(messages, alias[0], message)


if __name__ == "__main__":
    regist("test", "测试指令", alias=["test", "测试指令"])
    print(help_message("test"))
    print(help_message("支持"))



import inspect
from dataclasses import dataclass
from typing import Callable, Dict, Type, Tuple, Any, List

from multilogging import multilogger

from dicergirl.common import const
from dicergirl.common.const import DEFAULT_GROUP_NAME as NAME
from dicergirl.common.response import GenericResponse
from dicergirl.reply.data import GenericData, ConditionData
from dicergirl.reply.parsers.parser import parser
from dicergirl.reply.registry import ReplyRegistry


logger = multilogger(name="DicerGirl", payload="reply.manager")
""" `reply.manager`日志 """


@dataclass
class MethodInfo:
    """
    用于存储可调用对象及其参数信息的数据类

    Attributes:
        callable (Callable): 任意方法
        parameters (Dict[str, Type]): 参数信息的字典，其中键为参数名称，值为参数类型
    """

    callable: Callable
    parameters: Dict[str, Type]


class ReplyRegistryManager(ReplyRegistry):
    """
    ReplyRegistryManager 类用于管理全局方法、全局变量和处理消息事件等功能。

    Attributes:
        global_method (Dict[str, MethodInfo]): 用于存储全局方法的字典
        global_variable (Dict[str, Tuple[Type, Any]]): 用于存储全局变量的字典

    Methods:
        register_method: 注册全局方法
        remove_method: 删除全局方法
        register_variable: 注册全局变量
        remove_variable: 删除全局变量
        process_generic_event: 处理通用事件
        process_message_event: 处理消息事件
        call_method: 调用全局方法

    Private Methods:
        _handle_generic_event: 处理通用事件的内部方法
        _handle_placeholders: 处理占位符的内部方法
        _prepare_arguments: 准备方法参数的内部方法
        _filter_arguments: 过滤不匹配参数的内部方法
        _check_argument_types: 检查参数类型的内部方法
        _handle_condition_event: 处理条件事件的内部方法
        _execute_method: 执行方法的内部方法
    """

    _inspect_empty = inspect.signature(ReplyRegistry).empty
    global_method: Dict[str, MethodInfo] = {}
    global_variable: Dict[str, Tuple[Type, Any]] = {}

    def register_method(self, method: Callable, method_name: str = None):
        """
        注册全局方法

        Args:
            method (Callable): 待注册的方法
            method_name (str, optional): 方法的名称。如果未提供，默认使用方法的名称。
        """
        if method_name is None:
            method_name = method.__name__
        self.global_method[method_name] = MethodInfo(
            method,
            {
                name: parameter.annotation
                for name, parameter in inspect.signature(method).parameters.items()
            },
        )
        logger.debug(f"注册全局方法: {method}")

    def remove_method(self, method_name: str) -> bool:
        """
        删除全局方法

        Args:
            method_name (str): 待移除的方法名

        Returns:
            bool: 如果成功删除方法返回 True；否则返回 False。
        """
        if method_name in self.global_method:
            del self.global_method[method_name]
            logger.debug(f"销毁全局方法: {method_name}")
            return True
        return False

    def register_variable(self, **kwargs):
        """
        注册全局变量

        Args:
            **kwargs: 待注册的全局变量
        """
        for key, value in kwargs.items():
            self.global_variable[key] = (type(value), value)
            logger.debug(f"注册全局变量: {key}")

    def remove_variable(self, variable_name: str) -> bool:
        """
        删除全局变量

        Args:
            variable_name (str): 待删除的全局变量名

        Returns:
            bool: 如果成功删除变量，则返回 True；否则返回 False。
        """
        if variable_name in self.global_variable:
            logger.debug(f"销毁全局变量: {variable_name}")
            del self.global_variable[variable_name]
            return True
        return False

    def process_generic_event(self, event_name: str, **kwargs):
        """
        删除全局变量

        Args:
            event_name (str): 待处理的事件名
            **kwargs: 附加参数

        Returns:
            bool: 如果成功删除变量，则返回 True；否则返回 False。
        """
        for container in self._custom_generic_data.values():
            response = container.get_response(event_name)
            if response:
                is_enable = container.is_enable(event_name)
                if response and is_enable:
                    result = self._handle_generic_event(response, **kwargs)
                    if result:
                        return result
                    else:
                        break
        response = self._default_generic_data[NAME].get_response(event_name)
        result = self._handle_generic_event(response, **kwargs)
        if not result:
            logger.warning(f"{event_name}执行结果为: {result}")
        return result

    def process_message_event(self, message: str):
        """
        处理消息事件并返回匹配字段的消息文本

        Args:
            message (str): 用户发送的文本消息

        Returns:
            str | None: 匹配字段的消息文本，如果没有匹配则返回 None。
        """
        result = self._handle_condition_event(message)
        if len(result) == 0:
            return None
        return result

    def call_method(self, method_name: str, **kwargs):
        """
        调用全局方法。

        Args:
          method_name (str): 待调用的方法名
          **kwargs: 方法所需参数

        Returns:
          Any | None: 方法的执行结果，如果发生错误则返回 None。
        """
        method_info = self.global_method[method_name]
        if method_info is not None:
            kwargs = self._prepare_arguments(method_name, kwargs)
            if kwargs is not None:
                return self._execute_method(method_info.callable, kwargs)
        else:
            logger.warning(f"未找到全局方法 '{method_name}'。")
        return None

    def _handle_generic_event(self, response: GenericResponse, **kwargs) -> str | None:
        if not response:
            return None
        kwargs = self._handle_placeholders(response.send_text, **kwargs)
        return parser.replacement(response.send_text, **kwargs)

    def _handle_placeholders(self, send_text, *args, **kwargs):
        for placeholder in parser.get_placeholders(send_text):
            if placeholder not in kwargs:
                if placeholder in self.global_variable:
                    kwargs[placeholder] = self.global_variable[placeholder][1]
                elif placeholder in self.global_method:
                    kwargs[placeholder] = self.call_method(placeholder, **kwargs)
        return kwargs

    def _prepare_arguments(self, method_name, kwargs):
        method_info = self.global_method[method_name]
        parameters = method_info.parameters
        if set(kwargs.keys()) != set(parameters.keys()):
            kwargs = self._filter_arguments(parameters, kwargs)
        if set(kwargs.keys()) != set(parameters.keys()):
            logger.warning(
                f"方法 '{method_name}' 需要拥有参数 {list(parameters.keys())} ，"
                f"但仅提供了 {list(kwargs.keys())}。"
            )
            return None
        kwargs = self._check_argument_types(method_name, kwargs, parameters)
        return kwargs

    def _filter_arguments(self, parameters, kwargs):
        kwargs = {k: v for k, v in kwargs.items() if k in parameters}

        kwargs.update(
            {
                k: v[1]
                for k, v in self.global_variable.items()
                if k not in kwargs
                and k in parameters
                and (parameters[k] is v[0] or parameters[k] is self._inspect_empty)
            }
        )
        return kwargs

    def _check_argument_types(self, method_name, kwargs, parameters):
        for parameter_name, parameter in kwargs.items():
            if (
                not isinstance(parameter, parameters[parameter_name])
                and parameters[parameter_name] is not self._inspect_empty
            ):
                logger.warning(
                    f"方法 '{method_name}' 的参数 '{parameter_name}' 传递的类型错误。 "
                    f"期望的类型为：{parameters[parameter_name]}，但实际类型为：{type(parameter)}。"
                )
                return None
        return kwargs

    def _handle_condition_event(self, message: str) -> List[str]:
        tmp_dict = {}
        messages = []
        is_one_time_match = const.IS_ONE_TIME_MATCH
        for container in self._condition_specific_data.values():
            responses = container.get_responses(message)
            tmp_dict.update(responses)
        for container, response in tmp_dict.items():
            if container.is_enable(response.event_name):
                kwarg = self._handle_placeholders(response.send_text)
                result = parser.replacement(response.send_text, **kwarg)
                if result is not None:
                    messages.append(result)
                if is_one_time_match:
                    break
        return messages

    @staticmethod
    def _execute_method(method: Callable, kwargs):
        try:
            return method(**kwargs)
        except Exception as e:
            logger.warning(f"执行方法 '{method.__name__}' 时发生错误：{str(e)}")
            return None


manager = ReplyRegistryManager()



import re


class MessageParser:
    """
    消息解析类
    """

    def __init__(self):
        self.regex = "{([^{}]+)}"

    def replacement(self, text, **kwargs):
        """
        替换基本元素
        """

        def replace(match):
            key = match.group(1)
            return str(kwargs.get(key, match.group(0)))

        # replace方法的返回值有类型限定
        return re.sub(self.regex, replace, text)

    def get_placeholders(self, send_text):
        """
        提取消息中的{}内的文本
        """
        return re.findall(self.regex, send_text)


parser = MessageParser()



import httpx

origins = [
    "https://dicer.unvisitor.site/store/plugins.json",
    "https://unvisitor.gitee.io/dicer/store/plugins.json",
]


async def get_plugins():
    async with httpx.AsyncClient() as client:
        response = await client.get("https://dicer.unvisitor.site/store/plugins.json")
        result = response.json()

        if "official" in result.keys():
            official = result["official"]
        else:
            official = {}

        if "community" in result.keys():
            community = result["community"]
        else:
            community = {}

        return official, community


async def get_plugins_mixed():
    async with httpx.AsyncClient() as client:
        response = await client.get("https://dicer.unvisitor.site/store/plugins.json")
        result = response.json()

        if "official" in result.keys():
            official = result["official"]
        else:
            official = {}

        if "community" in result.keys():
            community = result["community"]
        else:
            community = {}

        official.update(community)
        return official


async def get_official_plugins():
    async with httpx.AsyncClient() as client:
        response = await client.get("https://dicer.unvisitor.site/store/plugins.json")
        result = response.json()

        if "official" in result.keys():
            official = result["official"]
        else:
            official = {}

        return official


async def get_community_plugins():
    async with httpx.AsyncClient() as client:
        response = await client.get("https://dicer.unvisitor.site/store/plugins.json")
        result = response.json()

        if "community" in result.keys():
            community = result["community"]
        else:
            community = {}

        return community


if __name__ == "__main__":
    import asyncio

    print(asyncio.run(get_plugins()))



from typing import Dict, List, Literal
from ..common.const import SUPERUSER_FILE
from .handlers import get_user_id

import uuid
import json


su_uuid: str
""" 超级管理员鉴权令牌 """


def make_uuid() -> str:
    """创建新的超级管理员鉴权令牌"""
    global su_uuid
    su_uuid = (str(uuid.uuid1()) + str(uuid.uuid4())).replace("-", "")
    return su_uuid


def get_uuid() -> str:
    """获取超级管理员鉴权令牌"""
    return su_uuid


def add_super_user(message) -> bool:
    """新增超级管理员"""
    with open(SUPERUSER_FILE, "w+") as _su:
        sr = _su.read()
        if not sr:
            sudos = {}
        else:
            sudos = json.loads(sr)
        sudos[get_user_id(message)] = ""
        _su.write(json.dumps(sudos))
    return True


def rm_super_user(message) -> bool:
    """删除超级管理员"""
    rsu = open(SUPERUSER_FILE, "r")
    sr = rsu.read()
    if not sr:
        return False
    sudos = json.loads(sr)
    try:
        sudos.pop(get_user_id(message))
    except KeyError:
        return False
    _su = open(SUPERUSER_FILE, "w")
    _su.write(json.dumps(sudos))
    return True


def is_super_user(event) -> bool:
    """判断`event`所指向的用户是否为超级管理员"""
    su = False
    with open(SUPERUSER_FILE, "r") as _su:
        sr = _su.read()
        if not sr:
            sudos = {}
        else:
            sudos = json.loads(sr)
    for sudo in sudos.keys():
        if get_user_id(event) == sudo:
            su = True
            break
    return su


def get_super_users() -> List[str]:
    """捕获所有骰娘管理员"""
    su_read = SUPERUSER_FILE.read_text()
    if not su_read:
        return []

    read: Dict[str, Literal[""]] = json.loads(su_read)
    return list(read.keys())



from ..utils.version import run_shell_command
from ..common.exceptions.pluginerror import (
    PluginNotFoundError,
    PluginExistsError,
    PluginInstallFailedError,
    PluginUninstallFailedError,
)
from ..utils.plugins import modes
from .parse import get_plugins_mixed
from multilogging import multilogger

import sys


logger = multilogger(name="DicerGirl", payload="plugins.operation")
""" `plugins.operation`日志 """


async def install(name):
    plugins = await get_plugins_mixed()

    if name not in plugins.keys():
        return PluginNotFoundError
    elif name in modes.keys():
        return PluginExistsError

    rsc = await run_shell_command(
        f"\"{sys.executable}\" -m pip install {plugins[name]['package']} -i https://pypi.org/simple"
    )
    if rsc["returncode"] != 0:
        logger.error(rsc["stderr"])
        return PluginInstallFailedError

    return True


async def remove(name):
    plugins = await get_plugins_mixed()

    if name in plugins.keys():
        package = plugins[name]["package"]
    else:
        package = name

    rsc = await run_shell_command(f'"{sys.executable}" -m pip uninstall {package} -y')

    if rsc["returncode"] != 0:
        logger.error(rsc["stderr"])
        return PluginUninstallFailedError

    return True


async def upgrade(name):
    plugins = await get_plugins_mixed()

    if name not in plugins.keys():
        return PluginNotFoundError

    rsc = await run_shell_command(
        f"\"{sys.executable}\" -m pip install {plugins[name]['package']} -i https://pypi.org/simple --upgrade"
    )
    if rsc["returncode"] != 0:
        logger.error(rsc["stderr"])
        return PluginInstallFailedError

    return True



from typing import Dict, Literal
from multilogging import multilogger

logger = multilogger(name="Dicer Girl", payload="Settings")
""" `settings.py`日志管理系统 """
status: Dict[str, bool] = {}
""" 机器人在各个群聊状态 """
DEBUG = False
""" 漏洞监测模式 """


def change_status(var) -> bool:
    """录入新的`status`内容"""
    global status
    status = var
    return True


def load_status_settings() -> bool:
    """导出当前机器人在各群聊的状态"""
    return status


def is_debug() -> Literal[False]:
    return DEBUG


def debugon() -> Literal[True]:
    global DEBUG
    DEBUG = True


def debugoff() -> Literal[True]:
    global DEBUG
    DEBUG = False



from nonebot.adapters.onebot.v11 import MessageEvent

from ..utils.dicer import Dicer
from ..utils.docimasy import judger
from ..utils.plugins import modes
from ..utils.cards import Cards
from ..utils.charactors import Character
from ..reply.manager import manager


class StatusCode:
    def __init__(self, status_code: int = 1) -> None:
        self.status_code = status_code

    def __eq__(self, __value: object) -> bool:
        return __value == self.status_code

    def __bool__(self):
        if self.status_code > 0:
            return True
        else:
            return False


def __set_plus_format(args: list):
    """`.set 技能 +x`语法解析"""
    while True:
        try:
            index = args.index("+")
        except:
            break
        args[index] = args[index] + args[index + 1]
        args.pop(index + 1)

    while True:
        try:
            index = args.index("-")
        except:
            break
        args[index] = args[index] + args[index + 1]
        args.pop(index + 1)

    return args


def __set_default(
    args: list,
    event: MessageEvent,
    reply: list,
    cards: Cards = None,
    module=None,
    attrs_dict: dict = None,
    cha: Character = None,
    qid: str = None,
) -> bool:
    """设置属性"""
    for attr, alias in attrs_dict.items():
        if args[0] in alias:
            if attr in ["名字", "性别"]:
                if attr == "性别" and not args[1] in ["男", "女"]:
                    reply.append(
                        manager.process_generic_event(
                            "BadSex", event=event, CharactorName="角色", Value=args[1]
                        )
                    )
                    return StatusCode(-1)
                cha.__dict__[alias[0]] = args[1]
            else:
                try:
                    if not args[1].startswith(("-", "+")):
                        cha.__dict__[alias[0]] = int(args[1])
                    elif args[1].startswith("+"):
                        cha.__dict__[alias[0]] += int(args[1][1:])
                    elif args[1].startswith("-"):
                        cha.__dict__[alias[0]] -= int(args[1][1:])
                except ValueError:
                    reply.append(
                        manager.process_generic_event(
                            "ValueError", event=event, SkillName=args[0], Value=args[1]
                        )
                    )
                    return StatusCode(0)
            cards.update(event, cha.__dict__, qid=qid)
            return StatusCode(1)


def __set_skill(
    args,
    event: MessageEvent,
    reply: list,
    cards: Cards = None,
    cha: Character = None,
    module=None,
    qid: str = None,
) -> bool:
    """设置技能"""
    try:
        if not args[1].startswith(("-", "+")):
            cha.skills[args[0]] = int(args[1])
        elif args[1].startswith("+"):
            cha.skills[args[0]] += int(args[1][1:])
        elif args[1].startswith("-"):
            cha.skills[args[0]] -= int(args[1][1:])
        cards.update(event, cha.__dict__, qid=qid)
        return StatusCode(1)
    except ValueError:
        reply.append(
            manager.process_generic_event(
                "ValueError", event=event, SkillName=args[0], Value=args[1]
            )
        )
        return StatusCode(0)


def set_handler(event: MessageEvent, args: list, at: list, mode: str = None) -> str:
    """兼容所有模式的`.set`指令后端方法"""
    module = modes[mode]
    cards: Cards = module.__cards__
    cache_cards: Cards = module.__cache__
    charactor: Character = module.__charactor__()
    attrs_dict: dict = module.__baseattrs__
    args: list = __set_plus_format(args)

    attr_saved = 0
    skill_saved = 0
    saved_failed = 0

    if len(at) == 1:
        qid = at[0]
    else:
        qid = ""

    if not args:
        if cache_cards.get(event, qid=qid):
            card_data = cache_cards.get(event, qid=qid)
            cards.update(event, card_data, qid=qid)
            cha = charactor.load(card_data)
            cache_cards.delete(event)
            return manager.process_generic_event(
                "CardSaved", event=event, CardDetail=cha.output()
            )
        else:
            return manager.process_generic_event(
                "CacheNotFound", event=event, ModuleName=module.__name__
            )
    else:
        if cards.get(event, qid=qid):
            card_data = cards.get(event, qid=qid)
            cha = charactor.load(card_data)
        else:
            return manager.process_generic_event(
                "CacheNotFound", event=event, ModuleName=module.__name__
            )

        if len(args) % 2 != 0:
            return manager.process_generic_event(
                "AttributeCountError", event=event, Command="set"
            )

        reply = []

        li = []
        sub_li = []
        for arg in args:
            index = args.index(arg)
            if index % 2 == 0:
                sub_li.append(arg)
            elif index % 2 == 1:
                sub_li.append(arg)
                li.append(sub_li)
                sub_li = []
            else:
                return manager.process_generic_event(
                    "AttributeCountError", event=event, Command="set"
                )

        for sub_li in li:
            set_default_code = __set_default(
                sub_li,
                event,
                reply,
                cards=cards,
                module=module,
                attrs_dict=attrs_dict,
                cha=cha,
                qid=qid,
            )
            if set_default_code:
                attr_saved += 1
                continue
            elif set_default_code == -1:
                saved_failed += 1
                continue

            if __set_skill(
                sub_li, event, reply, cards=cards, cha=cha, module=module, qid=qid
            ):
                skill_saved += 1
            else:
                saved_failed += 1

        if not saved_failed:
            return manager.process_generic_event(
                "OnSet",
                event=event,
                AttrSetNumber=attr_saved,
                SkillSetNumber=skill_saved,
                SkillSetFailed=saved_failed,
            )
        else:
            details = ""
            for detail in reply:
                details += detail + "\n"
            details.strip("\n")

            return manager.process_generic_event(
                "OnSetWithFailure",
                event=event,
                AttrSetNumber=attr_saved,
                SkillSetNumber=skill_saved,
                SkillSetFailed=saved_failed,
                FailedDetail=details,
            )


def show_handler(event: MessageEvent, args, at, mode=None):
    """兼容所有模式的`.show`指令后端方法"""
    module = modes[mode]
    cards: Cards = module.__cards__
    cache_cards: Cards = module.__cache__
    charactor: Character = module.__charactor__()
    attrs_dict: dict = module.__baseattrs__

    if len(at) == 1:
        qid = at[0]
    else:
        qid = ""

    reply = []
    if not args:
        if cards.get(event, qid=qid):
            card_data = cards.get(event, qid=qid)
            cha = charactor.load(card_data)
            reply.append(
                manager.process_generic_event(
                    "CardInUse", event=event, CardDetail=cha.output()
                )
            )

        if cache_cards.get(event, qid=qid):
            card_data = cache_cards.get(event, qid=qid)
            cha = charactor.load(card_data)
            reply.append(
                manager.process_generic_event(
                    "CardInCache", event=event, CardDetail=cha.output()
                )
            )
    elif args[0] in ["detail", "de", "details"]:
        if cards.get(event, qid=qid):
            card_data = cards.get(event, qid=qid)
            cha = charactor.load(card_data)
            reply.append(cha.skills_output())
    else:
        if cards.get(event, qid=qid):
            card_data = cards.get(event, qid=qid)
            cha = charactor.load(card_data)
            if hasattr(cha, "out_" + args[0]):
                try:
                    reply.append(getattr(cha, "out_" + args[0])())
                except:
                    reply.append("查询时出现异常, 可能你想要查询的内容不存在?")
            else:
                name = args[0]
                exp = 1
                for key, attrs in attrs_dict.items():
                    if args[0] in attrs:
                        exp = getattr(cha, attrs[0])
                        name = key
                        break

                for skill in cha.skills.keys():
                    if args[0] == skill:
                        exp = cha.skills[skill]
                        break

                reply.append(f"{name}: {exp}")

    if not reply:
        reply.append("未查询到保存或暂存信息.")

    return reply


def del_handler(event: MessageEvent, args: list, at: list, mode: str = None):
    """兼容所有模式的`.del`指令后端方法"""
    module = modes[mode]
    cache_cards: Cards = module.__cache__
    cards: Cards = module.__cards__

    if len(at) == 1:
        qid = at[0]
    else:
        qid = ""

    r = []
    for arg in args:
        if not arg:
            pass
        elif arg == "cache":
            if cache_cards.get(event, qid=qid):
                if cache_cards.delete(event, save=False):
                    r.append(
                        manager.process_generic_event(
                            "CacheCardCleared",
                            event=event,
                        )
                    )
                else:
                    r.append(
                        manager.process_generic_event(
                            "UnknownError",
                            event=event,
                        )
                    )
            else:
                r.append("暂无缓存人物卡数据.")
        elif arg == "card":
            if cards.get(event):
                if cards.delete(event):
                    r.append(
                        manager.process_generic_event(
                            "CardDeleted",
                            event=event,
                        )
                    )
                else:
                    r.append(
                        manager.process_generic_event(
                            "UnknownError",
                            event=event,
                        )
                    )
            else:
                r.append("暂无使用中的人物卡.")
        else:
            if cards.delete_skill(event, arg):
                r.append(
                    manager.process_generic_event(
                        "SkillDeleted", event=event, SkillName=arg
                    )
                )

    if not r:
        r.append("使用`.help del`获得指令使用帮助.")

    return r


def roll(event: MessageEvent, args: str, name: str = None) -> str:
    """标准掷骰指令后端方法"""
    time = 1
    if "#" in args:
        args = args.split("#")

        try:
            time = int(args[0].strip())
            if time > 20:
                return manager.process_generic_event(
                    "BadMultipleRollString", SenderCard=name
                )
        except ValueError:
            return manager.process_generic_event(
                "MultipleRollStringError", SenderCard=name
            )

        if len(args) == 1:
            args = "1d100"
        else:
            args = args[1]
    else:
        args = args.strip()

    args = args.split()
    if len(args) > 1:
        reason = args[1]
    else:
        reason = None

    args = args[0]

    try:
        dice = Dicer(args)
        r = judger(event, dice, None, name=name, reason=reason)

        for _ in range(time - 1):
            r += judger(event, dice, None, name=name, reason=reason)

        return r.detail
    except ValueError:
        return manager.process_generic_event("BadRollString", SenderCard=name)


def shoot(event: MessageEvent):
    dice = Dicer("1d20").roll()
    result = dice.outcome

    if result < 4:
        rstr = "右腿"
    elif result < 7:
        rstr = "左腿"
    elif result < 11:
        rstr = "腹部"
    elif result < 16:
        rstr = "胸部"
    elif result < 18:
        rstr = "右臂"
    elif result < 20:
        rstr = "左臂"
    elif result < 21:
        rstr = "头部"

    return manager.process_generic_event(
        "ShootDocimasy", event=event, DiceDescription=dice.description(), OnShoot=rstr
    )



import re


def require_update(old_version: str, new_version: str):
    regex = r"^(\d+)\.(\d+)\.(\d+)(.*?)?(\d+?)?$"

    old_tuple = re.match(regex, old_version)
    new_tuple = re.match(regex, new_version)

    old_tuple_main = tuple(map(int, filter(None, old_tuple.group(1, 2, 3))))
    new_tuple_main = tuple(map(int, filter(None, new_tuple.group(1, 2, 3))))

    new_length = len(list(filter(None, new_tuple.groups())))
    old_length = len(list(filter(None, old_tuple.groups())))

    if old_length == 5:
        old_pre_release = old_tuple.group(4)[0]
        old_pre_version = int(old_tuple.group(5))
    elif old_length == 3:
        old_pre_release = "s"
        old_pre_version = 1
    else:
        return False

    if new_length == 5:
        new_pre_release = new_tuple.group(4)[0]
        new_pre_version = int(new_tuple.group(5))
    elif new_length == 3:
        new_pre_release = "s"
        new_pre_version = 1
    else:
        return False

    if old_tuple_main < new_tuple_main:
        return True
    elif old_tuple_main > new_tuple_main:
        return False
    elif old_tuple_main == new_tuple_main:
        if old_pre_release < new_pre_release:
            return True
        elif old_pre_release > new_pre_release:
            return False
        elif old_pre_release == new_pre_release:
            if old_pre_version < new_pre_version:
                return True
            elif old_pre_version > new_pre_version:
                return False
            elif old_pre_version == new_pre_version:
                return False



from typing import List
from multilogging import multilogger

import re


logger = multilogger(name="DicerGirl", payload="utils.cards")
""" `utils.cards`日志 """


def translate_punctuation(string) -> str:
    """中文字符转换为英文字符"""
    punctuation_mapping = {
        "，": ",",
        "。": ".",
        "！": "!",
        "？": "?",
        "；": ";",
        "：": ":",
        "“": '"',
        "”": '"',
        "‘": "'",
        "’": "'",
        "（": "(",
        "）": ")",
        "【": "[",
        "】": "]",
        "《": "<",
        "》": ">",
    }
    for ch_punct, en_punct in punctuation_mapping.items():
        string = string.replace(ch_punct, en_punct)
    return string


def format_str(message: str, begin=None, lower=True) -> str:
    """骰娘指令转义及解析"""
    regex = r"[<\[](.*?)[\]>]"
    message = str(message).lower() if lower else str(message)
    msg = re.sub("\s+", " ", re.sub(regex, "", message)).strip(" ")
    msg = translate_punctuation(msg)
    logger.debug(msg)

    if begin:
        if isinstance(begin, str):
            begin = [
                begin,
            ]
        elif isinstance(begin, tuple):
            begin = list(begin)

        begin.sort(reverse=True)
        for b in begin:
            msg = msg.replace(b, "").lstrip(" ")

    logger.debug(msg)
    return msg


def format_msg(message, begin=None, zh_en=True) -> List[str]:
    """骰娘指令拆析为`list`的方法"""
    msgs = format_str(message, begin=begin)
    outer = []
    regex = (
        r'([+-]?\d+)|("[^"]+")|([a-zA-Z\u4e00-\u9fa5]+)'
        if not zh_en
        else r'([+-]?\d+)|([a-zA-Z]+)|("[^"]+")|([\u4e00-\u9fa5]+)'
    )
    msgs = list(filter(None, re.split(regex, msgs)))
    logger.debug(msgs)

    for msg in msgs:
        splited_msg = list(filter(None, re.split(regex, msg.strip(" "))))

        for i, msg in enumerate(splited_msg):
            splited_msg[i] = msg.strip('"')

        outer += splited_msg

    msgs = list(filter(None, outer))
    logger.debug(msgs)
    return msgs



from typing import List
from yaml.loader import FullLoader
from ..common.const import BLACKLIST_FILE

import yaml


class BlackList:
    """黑名单类"""

    def __init__(self) -> None:
        if not BLACKLIST_FILE.exists():
            BLACKLIST_FILE.parent.mkdir(parents=True, exist_ok=True)
            yaml.dump({"user": None, "group": None}, BLACKLIST_FILE.open("w"))

        if not self.load():
            raise OSError

    def load(self) -> bool:
        try:
            self.blacklist: list = (
                yaml.load(BLACKLIST_FILE.open("r"), FullLoader)["user"] or []
            )
            self.group_blacklist: list = (
                yaml.load(BLACKLIST_FILE.open("r"), FullLoader)["group"] or []
            )
            return True
        except:
            return False

    def dump(self) -> str:
        BLACKLIST_FILE.write_text("")
        return yaml.dump(
            {"user": self.blacklist, "group": self.group_blacklist},
            BLACKLIST_FILE.open("w"),
        )

    def get_blacklist(self) -> List[str]:
        return self.blacklist or []

    def add_blacklist(self, qid: str) -> str:
        self.blacklist.append(qid)
        return self.dump()

    def remove_blacklist(self, qid: str) -> str:
        self.blacklist.remove(qid)
        return self.dump()

    def get_group_blacklist(self) -> List[str]:
        return self.group_blacklist or []

    def add_group_blacklist(self, gid: str) -> str:
        self.group_blacklist.append(gid)
        return self.dump()

    def remove_group_blacklist(self, gid: str) -> str:
        self.group_blacklist.remove(gid)
        return self.dump()


blacklist = BlackList()



from ..common.const import BOT_MODES_FILE
from .handlers import get_group_id
from typing import Dict

import json


def load_modes() -> Dict[str, list]:
    """加载当前不同群聊的跑团模式"""
    return json.loads(open(BOT_MODES_FILE, "r").read())


def set_mode(event, mode) -> bool:
    """设置当前群聊的跑团模式"""
    lm = load_modes()
    lm[get_group_id(event)] = mode
    json.dump(lm, open(BOT_MODES_FILE, "w"))


def get_mode(event) -> str:
    """获得当前群聊的跑团模式"""
    lm = load_modes()
    if not get_group_id(event) in lm.keys():
        lm[get_group_id(event)] = "coc"
        json.dump(lm, open(BOT_MODES_FILE, "w"))
        return "coc"

    return lm[get_group_id(event)]



from ..utils.operator import get_name
from ..utils.handlers import get_group_id, get_user_card, get_user_id, get_user_nickname
from ..utils.modes import get_mode
from ..utils.operator import get_status
from ..reply.manager import manager
from .const import VERSION


def regist_all():
    regist_events()
    regist_vars()


def regist_vars():
    manager.register_method(get_name, "BotName")
    manager.register_method(get_group_id, "GroupID")
    manager.register_method(get_user_id, "UserID")
    manager.register_method(get_user_card, "SenderCard")
    manager.register_method(get_mode, "Mode")
    manager.register_method(get_status, "Status")
    manager.register_method(get_user_nickname, "SenderNickName")
    manager.register_variable(Version=VERSION)


def regist_events():
    regist_main_event()
    regist_general_event()
    regist_docimasy_event()


def regist_main_event():
    manager.register_event(
        "PermissionDenied", "[{SenderCard}]权限不足, {BotName}拒绝执行该指令.\n请先执行`.su`开启权限鉴定."
    )
    manager.register_event(
        "CommandPermissionDenied",
        "[{SenderCard}]权限不足, {BotName}拒绝执行指令`{Command}`.\n请先执行`.su`开启权限鉴定.",
    )
    manager.register_event("DebugOn", "[{SenderCard}]漏洞调试模式已启动.")
    manager.register_event("DebugOff", "[{SenderCard}]漏洞调试模式已关闭.")
    manager.register_event("NotManagerYet", "[{SenderCard}]你还不是超级管理员, 无法撤销超级管理员身份.")
    manager.register_event("AlreadyManager", "[{SenderCard}]你已经是超级管理员.")
    manager.register_event("ManagerExit", "[{SenderCard}]你已撤销超级管理员身份.")
    manager.register_event(
        "AuthenticateStarted", "[{SenderCard}]启动超级管理员鉴权, 鉴权令牌已在控制终端展示."
    )
    manager.register_event(
        "AuthenticateFailed",
        "[{SenderCard}]鉴权失败, 请检查你的令牌是否正确复制!\n使用`.help sudo`获取帮助信息.",
    )
    manager.register_event("AuthenticateSuccess", "[{SenderCard}]你成功取得了管理员权限.")
    manager.register_event("GroupLeaveSet", "{BotName}离开群聊, 期待与诸君的下一次见面.")
    manager.register_event("BotOn", "{BotName}已开放指令限制.")
    manager.register_event("BotOff", "{BotName}已开启指令限制.")
    manager.register_event("NameSet", "倒是好生有趣的名字, 以后我就是“{NewName}”了.")
    manager.register_event("UninstallFailed", "诶? 卸载失败?")
    manager.register_event("CommandFailed", "指令执行失败, 疑似模式[{Mode}]不存在该指令.")
    manager.register_event("UnknownMode", "未知的跑团模式[{Mode}].")
    manager.register_event("SetPermissionDenied", "仅允许主持人为其它玩家设置数据.")
    manager.register_event("DeletePermissionDenied", "仅允许主持人删除其它玩家数据.")
    manager.register_event("ModeChanged", "已切换跑团模式为[{Mode}].")
    manager.register_event("GroupApproval", "[{UserID}]邀请进入群[{GroupID}].\n处理方式: 同意.")
    manager.register_event("GroupForbidden", "[{UserID}]邀请进入群[{GroupID}].\n处理方式: 拒绝.")
    manager.register_event("BlacklistAdded", "用户[{UserID}]执行非法操作, 已列入黑名单.")
    manager.register_event(
        "CardInit", "[{SenderCard}]角色卡已参照模板[{Mode}]初始化, 异世的诗篇或许就要开始了."
    )
    manager.register_event("RoleOB", "[{SenderNickName}]身份组设置为旁观者 (OB).")
    manager.register_event("RoleKP", "[{SenderNickName}]身份组设置为主持人 (KP).")
    manager.register_event("main.friend.approve", "[{UserID}]请求添加好友.\n处理方式: 同意.")
    manager.register_event("main.friend.reject", "[{UserID}]请求添加好友.\n处理方式: 拒绝.")
    manager.register_event("main.test.userid", "[{UserID}]")
    manager.register_event(
        "main.friend.new",
        "欢迎使用 DicerGirl {Version}, 我是{BotName}. 使用`.help`查看帮助信息, 使用`.bot exit`移除群骰娘, 严禁踢出骰娘. 加入公测群[770386358]可以获得更多支持.",
    )
    manager.register_event("main.set.name", "[{SenderCard}]角色被命名为[{Name}], 新的诗篇就要被奏响了.")


def regist_general_event():
    manager.register_event(
        "OnSet", "[{SenderCard}]录卡完成, 成功设置基础属性{AttrSetNumber}个, 技能{SkillSetNumber}个."
    )
    manager.register_event(
        "OnSetWithFailure",
        "[{SenderCard}]录卡完成, 成功设置基础属性{AttrSetNumber}个, 技能{SkillSetNumber}个, 有{SkillSetFailed}个技能设置失败:\n{FailedDetail}",
    )
    manager.register_event("CardInUse", "[{SenderCard}]使用中人物卡: \n{CardDetail}")
    manager.register_event("CardInCache", "[{SenderCard}]已暂存人物卡: \n{CardDetail}")
    manager.register_event("CardSaved", "[{SenderCard}]成功从缓存保存人物卡属性: \n{CardDetail}")
    manager.register_event("CardDeleted", "[{SenderCard}]已删除使用中的人物卡！")
    manager.register_event("CacheCardCleared", "[{SenderCard}]已清空暂存人物卡数据.")
    manager.register_event(
        "BadRollString", "诶, 出错了, 请检查你的掷骰表达式.\n使用`.help roll`获得掷骰指令使用帮助."
    )
    manager.register_event("BadMultipleRollString", "[{SenderCard}]多轮掷骰次数过多, 指令拒绝.")
    manager.register_event("MultipleRollStringError", "参数错误, `#`提示符前应当跟随整型数.")
    manager.register_event(
        "BadSex", "{BotName}拒绝将{CharactorName}性别将设置为[{Value}], 这是对物种的侮辱."
    )
    manager.register_event(
        "AttributeCountError",
        "参数错误, 这是由于传输的数据数量错误, {BotName}只接受为偶数的参数数量, 这看起来不像是来源于我的错误.\n使用`.help {Command}`查看使用帮助.",
    )
    manager.register_event("ValueError", "[{SkillName}]需要一个整型数但你传入了[{Value}].")
    manager.register_event("UnknownError", "诶, 貌似发生了未知的错误?")
    manager.register_event("SkillDeleted", "[{SenderCard}]已删除技能[{SkillName}], 唔, 真是可惜.")
    manager.register_event(
        "ShootDocimasy", "[{SenderCard}]进行射击检定:\n{DiceDescription}\n检定命中了[{OnShoot}]."
    )
    manager.register_event(
        "CacheNotFound", "未找到缓存数据, 请先使用无参数的`.{ModuleName}`指令进行车卡生成角色卡, 或使用`.st`指令进行录卡."
    )


def regist_docimasy_event():
    manager.register_event(
        "utils.docimasy", "[{SenderCard}]掷骰: {DiceDiscription}."
    )
    manager.register_event(
        "utils.docimasy.reason",
        "[{SenderCard}]由于[{Reason}]掷骰: {DiceDescription}.",
    )
    manager.register_event(
        "utils.docimasy.skill",
        "[{SenderCard}]进行技能[{SkillName}:{Value}]检定: {DiceDescription}, 检定[{Docimasy}].",
    )



"""
存放常量的文件
"""
from pathlib import Path
from typing import Any, Dict
from ruamel.yaml import YAML


EMPTY_CHAR = EMPTY_STRING = ""
EMPTY_ARRAY = EMPTY_LIST = []
EMPTY_DICT = {}
NONE = NULL = None
ZERO = 0

VERSION = "3.4.20"
""" Dicer Girl 版本号 """
HOME_PATH = Path.home()
""" 主目录 """
PLUGINS_PATH = Path(__file__).resolve().parent.parent / "plugins"
DICERGIRL_DATA_PATH = HOME_PATH / ".dicergirl"
""" DicerGirl 数据文件夹路径 """
CURRENT_DIR = Path(__file__).resolve().parent
""" Dicer Girl 当前目录 """
SAVED_DATA_PATH = DICERGIRL_DATA_PATH / "data"
""" 人物卡数据存储文件夹 """
DICERGIRL_LOGS_PATH = DICERGIRL_DATA_PATH / "log"
""" 日志数据存储文件夹 """
BOT_STATUS_FILE = SAVED_DATA_PATH / "status.json"
""" 机器人启用状态存储文件 """
SUPERUSER_FILE = SAVED_DATA_PATH / "super_user.json"
""" 管理员鉴权文件 """
LOGGERS_CACHE_FILE = SAVED_DATA_PATH / "loggers.json"
""" 日志信息存储文件 """
BOT_MODES_FILE = SAVED_DATA_PATH / "modes.json"
""" 跑团模式存储文件 """
BLACKLIST_FILE = SAVED_DATA_PATH / "blacklist.yml"
""" 黑名单存储文件 """
DEFAULT_GROUP_NAME = "default"
""" 默认消息事件组名 """
REPLY_YAML = YAML()
""" `ruamel.yaml`的YAML对象实例化 """
REPLY_FOLDER_PATH = DICERGIRL_DATA_PATH / "reply"
""" DicerGirl 自定义回复文件夹路径 """
GENERIC_REPLY_FILE_PATH = REPLY_FOLDER_PATH / "dg-default.yml"
""" 自定义通用回复配置文件默认路径 """
CONDITION_SPECIFIC_REPLY_FILE_PATH = REPLY_FOLDER_PATH / "default.yml"
""" 特定条件下回复配置文件默认路径 """
EXAMPLE_CONDITION_SPECIFIC_REPLY_FILE_PATH = REPLY_FOLDER_PATH / "example.yml"
""" 特定条件下回复的示例配置文件路径 """
IS_ONE_TIME_MATCH = False
""" 自定义是否只匹配一次 """
GENERIC_REPLY_FILE_CACHE: Dict[str, Any] = {}
""" 自定义通用回复配置文件缓存 """
CONDITION_SPECIFIC_REPLY_FILE_CACHE: Dict[str, Any] = {}
""" 特定条件下回复配置文件缓存 """
CUSTOM_GENERIC_TEMPLATE = """\
enable: true
author: "默认"
# 版本号
version: 1.0
# 自定义功能描述
description: "默认文件"
# 自定义规则列表
items: []
"""
CONDITION_SPECIFIC_TEMPLATE = """\
enable: true
author: "默认"
# 版本号
version: 1.0
# 自定义功能描述
description: "默认文件"
# 自定义规则列表
items: []
"""
EXAMPLE_CONDITION_SPECIFIC_TEMPLATE = """\
# 是否启用
enable: true
# 作者
author: "佚名"
# 版本号
version: 1.0
# 自定义功能描述
description: "示例模板"
# 自定义规则列表
items: 
    - default.example:
        # 是否启用
        enable: false
        # 发送内容
        send_text: "你好，现在是北京时间{time}！"
        # 匹配字段
        match_field: "/example time"
        # EXACT_MATCH(完全匹配),PARTIAL_MATCH(部分匹配),REGEX_MATCH(正则匹配),FUNCTION_MATCH(方法匹配)
        match_type: EXACT_MATCH 
"""



from pathlib import Path
from multilogging import multilogger
from ..common.const import PLUGINS_PATH

import sys
import importlib

logger = multilogger(name="Dicer Girl", payload="Plugin")

loaded = False
""" 首次加载标识, 防止nb指令重复注册 """


def modules():
    global loaded
    if loaded:
        return modes

    modes_dict = {}
    library_dict = {}
    sys.path.append(PLUGINS_PATH.__str__())

    for folder in PLUGINS_PATH.iterdir():
        if Path(folder).is_dir() and (Path(folder) / "__init__.py").exists():
            try:
                module = importlib.import_module(folder.name)
            except Exception as error:
                logger.exception(error)
                logger.error(f"插件 {folder.name} 导入失败.")
                continue

            if not hasattr(module, "__type__"):
                continue

            if module.__type__ not in ("plugin", "library"):
                continue
            else:
                if module.__type__ == "plugin":
                    module_type = "插件"
                elif module.__type__ == "library":
                    module_type = "库"

            if hasattr(module, "__nbcommands__"):
                commands: dict = module.__nbcommands__
            else:
                commands = {}

            if commands and not hasattr(module, "__nbhandler__"):
                logger.error(f"{module_type} {folder.name} 配置异常, 导入失败.")
                continue

            if hasattr(module, "__nbhandler__"):
                handlers = module.__nbhandler__
            else:
                handlers = {}

            for command, handler in commands.items():
                try:
                    getattr(handlers, command)(getattr(handlers, handler))
                except AttributeError:
                    logger.error(
                        f"{module_type} {folder.name} 中 Nonebot2 指令配置异常, 导入失败."
                    )
                    continue
                except Exception as error:
                    logger.error("未知错误:")
                    logger.exception(error)

            if module.__type__ == "plugin":
                modes_dict[module.__name__] = module
            elif module.__type__ == "library":
                library_dict[module.__name__] = module

            logger.success(f"{module_type} {folder.name.upper()} 导入完成.")

    sys.path.pop(-1)
    loaded = True
    return modes_dict, library_dict


modules_dict, library_dict = modules()

modes: dict = {module.split(".")[-1]: modules_dict[module] for module in modules_dict}
""" 已导入的跑团模块 """
library: dict = {module.split(".")[-1]: library_dict[module] for module in library_dict}
